from discminer.plottools import (get_discminer_cmap, make_substructures,
                                 make_up_ax, mod_minor_ticks, mod_major_ticks,
                                 use_discminer_style, mod_nticks_cbars)
from discminer.rail import Rail, Contours

import numpy as np
import matplotlib.pyplot as plt
from scipy.signal import savgol_filter
from scipy.interpolate import interp1d

from astropy import units as u
from astropy.io import fits
import json
import copy

from utils import (init_data_and_model,
                   get_noise_mask,
                   get_1d_plot_decorators,
                   load_moments,
                   load_disc_grid,
                   add_parser_args,
                   make_1d_legend,
                   MEDIUM_SIZE)

from argparse import ArgumentParser

use_discminer_style()

parser = ArgumentParser(prog='plot radial profiles', description='Plot radial profiles from moments and residuals [velocity, linewidth, [peakintensity, peakint]?')
args = add_parser_args(parser, moment=True, kind=True, surface=True, writetxt=True, mask_minor=True, mask_major=True, Rinner=True, Router=True)

#**********************
#JSON AND PARSER STUFF
#**********************
with open('parfile.json') as json_file:
    pars = json.load(json_file)

meta = pars['metadata']
best = pars['best_fit']
custom = pars['custom']

Mstar = best['velocity']['Mstar']
vsys = best['velocity']['vsys']
vel_sign = best['velocity']['vel_sign']
Rout = best['intensity']['Rout']
incl = best['orientation']['incl']

gaps = custom['gaps']
rings = custom['rings']
Rmax = 1.1*Rout*u.au #Max model radius, 10% larger than disc Rout

clabel, clabel_res, clim0, clim0_res, clim1, clim1_res, unit = get_1d_plot_decorators(args.moment)

#*******************
#LOAD DATA AND MODEL
#*******************
datacube, model = init_data_and_model()

noise_mean, mask = get_noise_mask(datacube, thres=2)
vchannels = datacube.vchannels
model.make_model()

#*************************
#LOAD DISC GEOMETRY
R, phi, z = load_disc_grid()

R_s = R[args.surface]*u.m.to('au')
phi_s = phi[args.surface]

#*************************
#LOAD MOMENT MAPS
moment_data, moment_model, mtags = load_moments(args)
if mtags['surf']=='both':
    surf_ref = 'upper'
else:
    surf_ref = args.surface
tag_base = mtags['base']

#**************************
#MASK AND COMPUTE RESIDUALS
moment_data = np.where(mask, np.nan, moment_data)
moment_model = np.where(mask, np.nan, moment_model)
moment_residuals = moment_data - moment_model
    
if args.moment=='velocity':
    moment_residuals_abs = np.abs(moment_data-vsys) - np.abs(moment_model-vsys)

#**************************
#MAKE PLOTS
beam_au = datacube.beam_size.to('au').value
R_prof = np.arange(args.Rinner*beam_au, args.Router*Rout, beam_au/5)
xlim0, xlim1 = 0.5*R_prof[0], 1.05*R_prof[-1]

def writetxt(arr, tag=''):
    if tag!='': tag = '_'+tag
    arr = np.asarray(arr).T
    np.savetxt('radial_profile_%s%s.dat'%(args.moment, tag), arr, fmt='%.6f', header='R[au] attribute stddev')

def get_normalisation(mask_ang, component='z'):
    """
    mask_ang : +- mask around disc minor axis
    component : velocity component to get normalisation for    
    """
    if component=='phi':
        div_factor = 4*np.sin(np.pi/2 - np.radians(mask_ang)) * np.abs(np.sin(incl))/(2*np.pi - 4*np.radians(mask_ang))
    elif component=='z':
        div_factor = -np.cos(np.abs(incl))        
    else: return comp
    
    return div_factor

def make_savgol(prof):
    try:
        ysav = savgol_filter(prof, 5, 3)
        ysav_deriv = savgol_filter(prof, 5, 3, deriv=1)
    except np.linalg.LinAlgError:
        ysav = prof
        ysav_deriv = None
    return ysav, ysav_deriv 


if args.moment=='velocity':
    mask_ang = args.mask_minor #+-mask around disc minor axis
    #*******************
    #VELOCITY COMPONENTS
    #*******************    

    #VZ
    rail_vz = Rail(model, moment_residuals, R_prof)
    vel_z, vel_z_error = rail_vz.get_average(mask_ang=mask_ang, surface=surf_ref)
    div_factor_z = get_normalisation(mask_ang, component='z')

    vel_z /= div_factor_z
    vel_z_error /= div_factor_z 

    #DVPHI
    rail_phi = Rail(model, moment_residuals_abs, R_prof)
    vel_phi, vel_phi_error = rail_phi.get_average(mask_ang=mask_ang, surface=surf_ref)
    div_factor_phi = get_normalisation(mask_ang, component='phi')

    vel_phi /= div_factor_phi
    vel_phi_error /= div_factor_phi 

    #VPHI
    rail_phi = Rail(model, np.abs(moment_data-vsys), R_prof)
    vel_rot, _ = rail_phi.get_average(mask_ang=mask_ang, surface=surf_ref)
    vel_rot /= div_factor_phi
    vel_rot_error = vel_phi_error

    #VR
    mask_r = mask | (np.abs(phi_s) < np.radians(args.mask_major)) | (np.abs(phi_s) > np.radians(180-args.mask_major))
    moment_data_r = np.where(mask_r, np.nan, moment_data) 

    f_vp = interp1d(R_prof, vel_rot)
    f_vz = interp1d(R_prof, vel_z)
    R_interp = np.where((R_s>R_prof[0]) & (R_s<R_prof[-1]), R_s, R_prof[0])

    vr = -1/(np.sin(phi_s)*np.sin(incl)) * (moment_data_r - vsys - vel_sign*f_vp(R_interp)*np.cos(phi_s)*np.sin(incl) + f_vz(R_interp)*np.cos(incl))

    rail_vr = Rail(model, vr, R_prof)
    vel_r, vel_r_error = rail_vr.get_average(mask_ang=0.0, surface=surf_ref, av_func=np.nanmedian)

    #WRITE?
    if args.writetxt: writetxt([R_prof, vel_z, vel_z_error], tag='vz')
    if args.writetxt: writetxt([R_prof, vel_r, vel_r_error], tag='vr')
    if args.writetxt: writetxt([R_prof, vel_phi, vel_phi_error], tag='vphi')
    if args.writetxt: writetxt([R_prof, vel_rot, vel_phi_error], tag='rotationcurve')
    
    #PLOT 3D VELOCITIES
    fig, ax = plt.subplots(ncols=1, nrows=1, figsize=(14,4))
        
    ysav_phi, ysav_phi_deriv = make_savgol(vel_phi)
    ax.plot(R_prof, ysav_phi, c='dodgerblue', lw=3, label=r'$\Delta\upsilon_\phi$', zorder=12)
    ax.fill_between(R_prof, vel_phi+vel_phi_error, vel_phi-vel_phi_error, color='dodgerblue', alpha=0.15, zorder=9)

    ysav_z, ysav_z_deriv = make_savgol(vel_z)    
    ax.plot(R_prof, ysav_z, c='k', lw=3, label=r'$\upsilon_{\rm z}$', zorder=8)
    ax.fill_between(R_prof, vel_z+vel_z_error, vel_z-vel_z_error, color='k', alpha=0.15, zorder=8)
        
    ysav_r, ysav_r_deriv = make_savgol(vel_r)
    ax.plot(R_prof, ysav_r, c='#FFB000', lw=3, label=r'$\upsilon_{\rm R}$', zorder=7)
    ax.fill_between(R_prof, vel_r+vel_r_error, vel_r-vel_r_error, color='#FFB000', alpha=0.15, zorder=7)

    ax.axhline(0, lw=2, ls='--', color='0.7')
    
    ax.set_xlabel('Radius [au]')
    ax.set_xlim(xlim0, xlim1)
    ax.set_ylabel(r'$\delta\upsilon$ [km/s]')
    ax.set_ylim(clim0_res, clim1_res)
    mod_major_ticks(ax, axis='x', nbins=10)
    mod_minor_ticks(ax)
    make_1d_legend(ax, fontsize=MEDIUM_SIZE+1)
    
    make_substructures(ax, gaps=gaps, rings=rings, label_gaps=True, label_rings=True)
    
    plt.savefig('velocity_components_%s.png'%tag_base, bbox_inches='tight', dpi=200)
    plt.show()

    #*******************
    #ROTATION CURVE
    #*******************
    fig, ax = plt.subplots(ncols=1, nrows=1, figsize=(14,4))        

    #DATA CURVE
    ysav_rot, ysav_rot_deriv = make_savgol(vel_rot)
    ax.plot(R_prof, ysav_rot, c='tomato', lw=3.5, label=r'Data', zorder=12)
    ax.fill_between(R_prof, vel_rot+vel_rot_error, vel_rot-vel_rot_error, color='tomato', alpha=0.15, zorder=9)
    
    #r"""
    #MODEL CURVE
    rail_phi = Rail(model, np.abs(moment_model-vsys), R_prof)
    vel_phi, _ = rail_phi.get_average(mask_ang=mask_ang, surface=surf_ref)
    vel_phi /= div_factor_phi
    ysav_phi, ysav_phi_deriv = make_savgol(vel_phi)
    ax.plot(R_prof, ysav_phi, c='0.7', lw=4.0, label=r'Model', zorder=11)
    #"""
    
    #PERFECT KEPLERIAN
    coords = {'R': R_prof*u.au.to('m')}
    velocity_upper = model.get_attribute_map(coords, 'velocity', surface=surf_ref) * vel_sign
    ax.plot(R_prof, velocity_upper, c='k', lw=2.5, ls='--', label=r'Keplerian (%.2f Msun)'%Mstar, zorder=13)

    #DECORATIONS
    ax.axhline(0, lw=2, ls='--', color='0.7')

    ax.set_xlabel('Radius [au]')
    ax.set_xlim(xlim0, xlim1)    
    ax.set_ylabel(r'Rotation velocity [km/s]')
    ax.set_ylim(clim0, 1.2*np.nanmax(vel_phi))
    mod_major_ticks(ax, axis='x', nbins=10)
    mod_minor_ticks(ax)    
    make_1d_legend(ax)

    make_substructures(ax, gaps=gaps, rings=rings, label_gaps=True, label_rings=True)  
    
    plt.savefig('rotation_curve_%s.png'%tag_base, bbox_inches='tight', dpi=200)
    plt.show()

    """
    #Data rotation curve - perfect_Keplerian
    fig, ax = plt.subplots(ncols=1, nrows=1, figsize=(14,4)) 
    ax.plot(R_prof, vel_rot-velocity_upper, c='k', lw=2.5, ls='--', label=r'Keplerian (%.2f Msun)'%Mstar, zorder=13)
    plt.show()
    """
    
else:
    mask_ang = 0.0
    #*****************
    #ABSOLUTE PROFILES
    #*****************
    kw_avg = dict(surface=surf_ref, av_func=np.nanmedian, mask_ang=mask_ang)
    
    fig, ax = plt.subplots(ncols=1, nrows=1, figsize=(14,4))

    #DATA CURVE
    rail_phi = Rail(model, moment_data, R_prof)
    vel_phi, vel_phi_error = rail_phi.get_average(**kw_avg)

    ysav_phi, ysav_phi_deriv = make_savgol(vel_phi)    
    ax.plot(R_prof, ysav_phi, c='tomato', lw=4.0, label=r'Data', zorder=12)
    ax.fill_between(R_prof, vel_phi+vel_phi_error, vel_phi-vel_phi_error, color='tomato', alpha=0.15, zorder=9)

    if args.writetxt: writetxt([R_prof, vel_phi, vel_phi_error], tag='data')

    #MODEL CURVE
    rail_phi = Rail(model, moment_model, R_prof)
    vel_phi, vel_phi_error = rail_phi.get_average(**kw_avg)

    ysav_phi, ysav_phi_deriv = make_savgol(vel_phi)    
    ax.plot(R_prof, ysav_phi, c='dodgerblue', lw=3.5, label=r'Model', zorder=11)

    if args.writetxt: writetxt([R_prof, vel_phi, vel_phi_error], tag='model')
    
    #DECORATIONS
    ax.axhline(0, lw=2, ls='--', color='0.7')

    ax.set_xlabel('Radius [au]')
    ax.set_xlim(xlim0, xlim1)    
    ax.set_ylabel(clabel)
    ax.set_ylim(clim0, clim1)
    mod_major_ticks(ax, axis='x', nbins=10)
    mod_minor_ticks(ax)
    make_1d_legend(ax)

    make_substructures(ax, gaps=gaps, rings=rings, label_gaps=True, label_rings=True)  
    
    plt.savefig('radial_profile_%s.png'%tag_base, bbox_inches='tight', dpi=200)
    plt.show()

    #****************
    #RESIDUALS
    #****************    
    fig, ax = plt.subplots(ncols=1, nrows=1, figsize=(14,4))

    rail_phi = Rail(model, moment_residuals, R_prof)
    vel_phi, vel_phi_error = rail_phi.get_average(**kw_avg)

    ysav_phi, ysav_phi_deriv = make_savgol(vel_phi)    
    ax.plot(R_prof, ysav_phi, c='tomato', lw=4.0, label=r'Residuals', zorder=12)
    ax.fill_between(R_prof, vel_phi+vel_phi_error, vel_phi-vel_phi_error, color='tomato', alpha=0.15, zorder=9)

    if args.writetxt: writetxt([R_prof, vel_phi, vel_phi_error], tag='residuals')    

    #DECORATIONS
    ax.axhline(0, lw=2, ls='--', color='0.7')

    ax.set_xlabel('Radius [au]')
    ax.set_xlim(xlim0, xlim1)    
    ax.set_ylabel(clabel_res)
    ax.set_ylim(clim0_res, clim1_res)
    mod_major_ticks(ax, axis='x', nbins=10)
    mod_minor_ticks(ax)

    make_substructures(ax, gaps=gaps, rings=rings, label_gaps=True, label_rings=True)
        
    plt.savefig('radial_profile_residuals_%s.png'%tag_base, bbox_inches='tight', dpi=200)
    plt.show()
    




import os
import sys
import numpy as np
from astropy import units as u
from astropy import constants as apc
from .. import constants as sfc
from .. import units as sfu

hypot_func = lambda x,y: np.sqrt(x**2 + y**2) #Slightly faster than np.hypot<np.linalg.norm<scipydistance. Checked precision up to au**2 orders.

class InputError(Exception):
    """Exception raised for input errors.

    Parameters
    ----------
    expression : str
        Input expression where error occurred
    
    message : str
        Output description of the error
    """
    def __init__(self, expression, message):
        self.expression = expression
        self.message = message
        
    def __str__(self):
        return '%s --> %s'%(self.expression, self.message)

    
class FrontendUtils(object):
    """
    Make outputs prettier
    """    
    path_icons = os.path.dirname(os.path.realpath(__file__))+'/../icons/'

    @staticmethod
    def _progress_bar(percent=0, width=50):
        left = width * percent // 100 
        right = width - left
        """
        print('\r[', '#' * left, ' ' * right, ']',
              f' {percent:.0f}%',
              sep='', end='', flush=True)
        """
        print('\r[', '#' * left, ' ' * right, ']', ' %.0f%%'%percent, sep='', end='') #compatible with python2 rtdocs
        sys.stdout.flush()

    @staticmethod
    def _break_line(init='', border='*', middle='=', end='\n', width=100):
        print('\r', init, border, middle * width, border, sep='', end=end)

    @staticmethod
    def _print_logo(filename=path_icons+'logo.txt'):
        logo = open(filename, 'r')
        print(logo.read())
        logo.close()
        

class FITSUtils(object):
    """
    Useful functions to transform datacube properties.
    """
    @staticmethod
    def _convert_to_tb(data, header, beam, planck=True, writefits=True, tag=""):
        """
        Convert intensity to brightness temperature in units of Kelvin.

        Parameters
        ----------
        data : (nchan, nx, nx) array_like
            Input intensity.
        
        header : FITS header
            Header of the datacube.
        
        beam : `~radio_beam.Beam` instance
            Beam from header, extracted with `~radio_beam`.
        
        planck : bool, optional
            If True, it uses the full Planck law to make the conversion, else it uses the Rayleigh-Jeans approximation. 

        writefits : bool, optional
            If True it creates a FITS file with the new intensity data and header.
        
        tag : str, optional
            String to add at the end of the output filename.

        kwargs : keyword arguments
            Additional keyword arguments to pass to `~astropy.io.fits.writeto` function.
           
        """
        I = data * u.Unit(header["BUNIT"]).to("beam-1 Jy")
        nu = header["RESTFRQ"]  # in Hz
        bmaj = beam.major.to(u.arcsecond).value
        bmin = beam.minor.to(u.arcsecond).value
        # area of gaussian beam
        beam_area = u.au.to("m") ** 2 * np.pi * (bmaj * bmin) / (4 * np.log(2))
        # beam solid angle: beam_area/(dist*pc)**2.
        #  dist**2 cancels out with beamarea's dist**2 from conversion of bmaj, bmin to mks units.
        beam_solid = beam_area / u.pc.to("m") ** 2
        Jy_to_SI = 1e-26
        c_h = apc.h.value
        c_c = apc.c.value
        c_k_B = apc.k_B.value

        if planck:
            Tb = (
                np.sign(I)
                * (
                    np.log(
                        (2 * c_h * nu ** 3)
                        / (c_c ** 2 * np.abs(I) * Jy_to_SI / beam_solid)
                        + 1
                    )
                )
                ** -1
                * c_h
                * nu
                / (c_k_B)
            )
        else:
            wl = c_c / nu
            Tb = 0.5 * wl ** 2 * I * Jy_to_SI / (beam_solid * c_k_B)

        return Tb


def get_tb(I, nu, beam, full=True):
    """
    nu in GHz
    Intensity in mJy/beam
    beam object from radio_beam
    if full: use full Planck law, else use rayleigh-jeans approximation
    """
    bmaj = beam.major.to(u.arcsecond).value
    bmin = beam.minor.to(u.arcsecond).value
    beam_area = sfu.au**2*np.pi*(bmaj*bmin)/(4*np.log(2)) #area of gaussian beam
    #beam solid angle: beam_area/(dist*pc)**2. dist**2 cancels out with beamarea's dist**2 from conversion or bmaj, bmin to mks units. 
    beam_solid = beam_area/sfu.pc**2 
    mJy_to_SI = 1e-3*1e-26
    nu = nu*1e9
    if full:
        Tb = np.sign(I)*(np.log((2*sfc.h*nu**3)/(sfc.c**2*np.abs(I)*mJy_to_SI/beam_solid)+1))**-1*sfc.h*nu/(sfc.kb) 
    else:
        wl = sfc.c/nu 
        Tb = 0.5*wl**2*I*mJy_to_SI/(beam_solid*sfc.kb) 
    #(1222.0*I/(nu**2*(beam.minor/1.0).to(u.arcsecond)*(beam.major/1.0).to(u.arcsecond))).value #nrao RayJeans             
    return Tb


def _get_beam_from(beam, dpix=None, distance=None, frac_pixels=1.0):
    """
    beam must be str pointing to fits file to extract beam from header or radio_beam Beam object.
    If radio_beam Beam instance is provided, pixel size (in SI units) will be extracted from grid obj. Distance (in pc) must be provided.
    #frac_pixels: number of averaged pixels on the data (useful to reduce computing time)
    """
    from astropy.io import fits
    from radio_beam import Beam
    sigma2fwhm = np.sqrt(8*np.log(2))
    if isinstance(beam, str):
        header = fits.getheader(beam)
        beam = Beam.from_fits_header(header)
        pix_scale = header['CDELT2'] * u.Unit(header['CUNIT2']) * frac_pixels
    elif isinstance(beam, Beam):
        if distance is None: raise InputError(distance, 'Wrong input distance. Please provide a value for the distance (in pc) to transform grid pix to arcsec')
        pix_radians = np.arctan(dpix / (distance*sfu.pc)) #dist*ang=projdist
        pix_scale = (pix_radians*u.radian).to(u.arcsec)  
    else: raise InputError(beam, 'beam object must either be str or Beam instance')

    x_stddev = ((beam.major/pix_scale) / sigma2fwhm).decompose().value 
    y_stddev = ((beam.minor/pix_scale) / sigma2fwhm).decompose().value 
    angle = (90*u.deg+beam.pa).to(u.radian).value
    gauss_kern = Gaussian2DKernel(x_stddev, y_stddev, angle) 
    
    #gauss_kern = beam.as_kernel(pix_scale) #as_kernel() slows down the run when passed to astropy.convolve
    return beam, gauss_kern


def weighted_std(prop, weights, weighted_mean=None):
    sum_weights = np.sum(weights)
    if weighted_mean is None:
        weighted_mean = np.sum(weights*prop)/sum_weights
    n = np.sum(weights>0)
    w_std = np.sqrt(np.sum(weights*(prop-weighted_mean)**2)/((n-1)/n * sum_weights))
    return w_std


def read_if_file_exists(base_str, file_str, n_none=3):
    if base_str is None: return [None]*n_none
    filename = base_str+file_str
    try:
        return np.loadtxt(filename)
    except OSError:
        message = 'File not found: %s'%filename
        warnings.warn(message)
        return [None]*n_none



from discminer.core import Data
from discminer.rail import Contours
from discminer.plottools import (add_cbar_ax,
                                 make_substructures,
                                 _make_radial_grid_2D,
                                 make_up_ax,
                                 mod_minor_ticks,
                                 mod_major_ticks,
                                 use_discminer_style)

from utils import (get_2d_plot_decorators,
                   get_noise_mask,
                   load_moments,
                   load_disc_grid,
                   add_parser_args)

import numpy as np
import matplotlib.pyplot as plt
from astropy import units as u

import json
import copy
from argparse import ArgumentParser

use_discminer_style()

parser = ArgumentParser(prog='plot moment maps', description='Plot moment map [velocity, linewidth, [peakintensity, peakint]?')
parser.add_argument('-c', '--coords', default='sky', type=str, choices=['disc', 'sky'], help="reference frame")
args = add_parser_args(parser, kind=True, surface=True)
     
#**********************
#JSON AND PARSER STUFF
#**********************
with open('parfile.json') as json_file:
    pars = json.load(json_file)

meta = pars['metadata']
best = pars['best_fit']
custom = pars['custom']

Rout = best['intensity']['Rout']
incl = best['orientation']['incl']
PA = best['orientation']['PA']
xc = best['orientation']['xc']
yc = best['orientation']['yc']

gaps = custom['gaps']
rings = custom['rings']

#****************
#SOME DEFINITIONS
#****************
file_data = meta['file_data']
tag = meta['tag']
au_to_m = u.au.to('m')

dpc = meta['dpc']*u.pc
Rmax = 1.1*Rout*u.au

#********************
#LOAD DATA AND GRID
#********************
datacube = Data(file_data, dpc) # Read data and convert to Cube object
noise_mean, mask = get_noise_mask(datacube)

#Useful definitions for plots
with open('grid_extent.json') as json_file:
    grid = json.load(json_file)

xmax = grid['xsky'] 
xlim = 1.15*np.min([xmax, Rmax.value]) 
extent= np.array([-xmax, xmax, -xmax, xmax])

#*************************
#LOAD DISC GEOMETRY
R, phi, z = load_disc_grid()

R_nonan_au = np.nan_to_num(R[args.surface])/au_to_m
z_nonan_au = np.nan_to_num(z[args.surface])/au_to_m

Xproj = R_nonan_au*np.cos(phi[args.surface])
Yproj = R_nonan_au*np.sin(phi[args.surface])

nh = int(0.5*datacube.nx)
#**********************************
#LOAD MOMENT MAPS AND GET RESIDUALS
moments_data = {}
moments_model = {}
residuals = {}
mtags = {}
mtypes = ['linewidth', 'velocity', 'peakintensity']

for moment in mtypes:
    md, mm, mt = load_moments(args, moment=moment)
    md = np.where(mask, np.nan, md)
    mm = np.where(mask, np.nan, mm)     
    mtags[moment] = mt
    moments_data[moment] = md 
    moments_model[moment] = mm 
    residuals[moment] = md - mm
         
#****************
#USEFUL FUNCTIONS
#****************
def clip_prop_radially(prop2d, Rmin=datacube.beam_size.to('au').value, Rmax=np.inf, Rgrid=R_nonan_au):
    return np.where((Rgrid<Rmin) | (Rgrid>Rmax), np.nan, prop2d)

def decorate_ax_res_2D(ax, cbar, lim=xlim):
    ax.set_aspect(1)
    make_up_ax(ax, xlims=(-lim,lim), ylims=(-lim,lim), labelsize=13)
    ax.set_xlabel('Offset [au]')
    ax.xaxis.set_label_position('top')    
    cbar.ax.tick_params(which='major', direction='in', width=2.7, size=4.8, pad=4, labelsize=12)
    cbar.ax.tick_params(which='minor', direction='in', width=2.3, size=3.3)
    mod_minor_ticks(cbar.ax)

#****************
#PLOT RESIDUALS
#****************    
fig, ax = plt.subplots(ncols=3, nrows=1, figsize=(17,6))
cbar_axes = [add_cbar_ax(fig, axi) for axi in ax]

clabels = {
    'linewidth': r'$\Delta$ Linewidth [km s$^{-1}$]',
    'velocity': r'$\Delta$ Centroid Velocity [km s$^{-1}$]',
    'peakintensity': r'$\Delta$ Peak Intensity [K]'
}
fill_angs_2pi = np.linspace(0, 2*np.pi, 100)

for i, (axi, moment) in enumerate(zip(ax, mtypes)):
    ctitle, clabel, clim, cfmt, cmap_mom, cmap_res, levels_im, levels_cc, unit = get_2d_plot_decorators(moment)
    levels_resid = np.linspace(-clim, clim, 32)
    levels_cbar = np.linspace(-clim, clim, 5)

    kwargs_im = dict(cmap=cmap_res, levels=levels_resid, extend='both', origin='lower')
    kwargs_cbar = dict(orientation='horizontal', format=cfmt, ticks=levels_cbar, pad=0.03, shrink=0.95, aspect=15)    
    
    if args.coords=='sky':
        #clip_prop_radially(residuals[moment])
        im = axi.contourf(residuals[moment], extent=extent, **kwargs_im)
        Contours.emission_surface(axi, R, phi, extent=extent,
                                  R_lev=np.linspace(0.1, 1.0, 10)*Rout*au_to_m,
                                  which=mtags['velocity']['surf']
        )
        Contours.disc_axes(axi,
                           R[args.surface][nh]/au_to_m,
                           z[args.surface][nh]/au_to_m,
                           incl, PA, xc=xc, yc=yc)
        
    elif args.coords=='disc':
        im = axi.contourf(Xproj, Yproj, clip_prop_radially(residuals[moment], Rmax=Rout), **kwargs_im)                         
        make_substructures(axi, gaps=gaps, rings=rings, twodim=True)
        _make_radial_grid_2D(axi, Rout, gaps=gaps, rings=rings, make_labels=True, label_freq=2)
        
    cbar = plt.colorbar(im, cax=cbar_axes[i], **kwargs_cbar)
    cbar.set_label(clabels[moment], fontsize=14)

    decorate_ax_res_2D(axi, cbar)
    datacube.plot_beam(axi, fc='lime')

ax[0].set_ylabel('Offset [au]')

tag_base = mtags['velocity']['base'].split('velocity_')[-1]
plt.savefig('residuals_all_%s_%sframe.png'%(tag_base, args.coords), bbox_inches='tight', dpi=200)
plt.show()
plt.close()



from discminer.core import Data

import json
from astropy import units as u
from argparse import ArgumentParser

parser = ArgumentParser(prog='make moments', description='Make gauss or bell moments and save outputs into .fits files')
parser.add_argument('-k', '--kind', default='gaussian', type=str, choices=['gauss', 'gaussian', 'bell'], help="1d moment kernel")
args = parser.parse_args()

with open('parfile.json') as json_file:
    pars = json.load(json_file)

meta = pars['metadata']

#****************
#SOME DEFINITIONS
#****************
dpc = meta['dpc']*u.pc

#*********
#LOAD DATA
#*********
tag = meta['tag']
file_data = 'cube_data_%s_convtb.fits'%tag
file_model = 'cube_model_%s_convtb.fits'%tag

datacube = Data(file_data, dpc) # Read data and convert to Cube object
modelcube = Data(file_model, dpc) # Read data and convert to Cube object

#**********************
#MAKE MOMENT MAPS
#**********************
moments_data = datacube.make_moments(method=args.kind, tag='data')
moments_model = modelcube.make_moments(method=args.kind, tag='model')




from discminer.core import Data
from discminer.rail import Contours
from discminer.plottools import (make_up_ax,
                                 mod_major_ticks,
                                 mod_nticks_cbars,
                                 use_discminer_style)

from utils import (get_2d_plot_decorators,
                   get_noise_mask,
                   load_moments,
                   load_disc_grid,
                   add_parser_args)

import numpy as np
import matplotlib.pyplot as plt
from astropy import units as u

import json
from argparse import ArgumentParser

use_discminer_style()

parser = ArgumentParser(prog='plot moment maps', description='Plot moment map [velocity, linewidth, [peakintensity, peakint]?')
args = add_parser_args(parser, moment=True, kind=True, surface=True)
     
#**********************
#JSON AND PARSER STUFF
#**********************
with open('parfile.json') as json_file:
    pars = json.load(json_file)

meta = pars['metadata']
best = pars['best_fit']
custom = pars['custom']
vsys = best['velocity']['vsys']
Rout = best['intensity']['Rout']

ctitle, clabel, clim, cfmt, cmap_mom, cmap_res, levels_im, levels_cc, unit = get_2d_plot_decorators(args.moment)

#****************
#SOME DEFINITIONS
#****************
file_data = meta['file_data']
tag = meta['tag']

dpc = meta['dpc']*u.pc
Rmax = 1.1*Rout*u.au #Max model radius, 10% larger than disc Rout

#********************
#LOAD DATA AND GRID
#********************
datacube = Data(file_data, dpc) # Read data and convert to Cube object
noise_mean, mask = get_noise_mask(datacube)

#Useful definitions for plots
with open('grid_extent.json') as json_file:
    grid = json.load(json_file)

xmax = grid['xsky'] 
xlim = 1.15*np.min([xmax, Rmax.value])
extent= np.array([-xmax, xmax, -xmax, xmax])

#*************************
#LOAD DISC GEOMETRY
R, phi, z = load_disc_grid()

#*************************
#LOAD MOMENT MAPS    
moment_data, moment_model, mtags = load_moments(args)
        
#**************************
#MASK AND COMPUTE RESIDUALS
moment_data = np.where(mask, np.nan, moment_data)
moment_model = np.where(mask, np.nan, moment_model)
residuals = moment_data - moment_model

"""
from scipy.ndimage import gaussian_filter
moment_data = gaussian_filter(moment_data, sigma=1.5)
moment_model = gaussian_filter(moment_model, sigma=1.5)
residuals = moment_data - moment_model
"""

#**************************
#MAKE PLOT

fig, ax = plt.subplots(ncols=3, nrows=1, figsize=(15,6))
ax_cbar0 = fig.add_axes([0.15, 0.14, 0.450, 0.04])
ax_cbar2 = fig.add_axes([0.68, 0.14, 0.212, 0.04])
    
kwargs_im = dict(cmap=cmap_mom, extent=extent, levels=levels_im)
kwargs_cc = dict(colors='k', linestyles='-', extent=extent, levels=levels_cc, linewidths=0.4)
kwargs_cbar = dict(orientation='horizontal', pad=0.03, shrink=0.95, aspect=15)

im0 = ax[0].contourf(moment_data, extend='both', **kwargs_im)
im1 = ax[1].contourf(moment_model, extend='both', **kwargs_im)
im2 = ax[2].contourf(residuals, cmap=cmap_res, origin='lower', extend='both', extent=extent, levels=np.linspace(-clim, clim, 32))

cc0 = ax[0].contour(moment_data, **kwargs_cc)
cc1 = ax[1].contour(moment_model, **kwargs_cc)

cbar0 = plt.colorbar(im0, cax=ax_cbar0, format='%.1f', **kwargs_cbar)
cbar0.ax.tick_params(labelsize=12) 
cbar2 = plt.colorbar(im2, cax=ax_cbar2, format=cfmt, **kwargs_cbar)
cbar2.ax.tick_params(labelsize=12) 

mod_nticks_cbars([cbar0], nbins=10)
mod_nticks_cbars([cbar2], nbins=5)

ax[0].set_ylabel('Offset [au]', fontsize=15)
ax[0].set_title(ctitle, pad=40, fontsize=17)
ax[1].set_title('Discminer Model', pad=40, fontsize=17)
ax[2].set_title('Residuals', pad=40, fontsize=17)

cbar0.set_label(clabel, fontsize=14)
cbar2.set_label(r'Residuals %s'%unit, fontsize=14)

for axi in ax:
    #axi.scatter(best['orientation']['xc'], best['orientation']['yc'], c='k', marker='X', lw=0.5, s=30) #use +offset.py script
    make_up_ax(axi, xlims=(-xlim, xlim), ylims=(-xlim, xlim), labelsize=11)
    mod_major_ticks(axi, axis='both', nbins=8)
    datacube.plot_beam(axi, fc='lime')
    axi.set_aspect(1)
    
for axi in ax[1:]: axi.tick_params(labelleft=False)

for i,axi in enumerate(ax):
    Contours.emission_surface(axi, R, phi, extent=extent, R_lev=np.linspace(0.1, 1.0, 10)*Rout*u.au.to('m'), which=mtags['surf'])
                              
    
plt.savefig('moment+residuals_%s.png'%mtags['base'], bbox_inches='tight', dpi=200)
plt.show()
plt.close()



"""
plottools module
===========
"""
import os
import copy
import matplotlib
import numpy as np
from math import ceil
import matplotlib.pyplot as plt
import matplotlib.patches as patches
from collections.abc import Iterable
from .tools.utils import weighted_std
from .tools.fit_kernel import _gauss
from .grid import GridTools
        
SMALL_SIZE = 10
MEDIUM_SIZE = 15
BIGGER_SIZE = 22

_residuals_colors = {
    'velocity': ["#"+tmp for tmp in ["000b14","004f8f","1f9aff","fff7db","ff5a47","cc0033","140200"]],
    'linewidth': ["#"+tmp for tmp in ["000b14","036407","81ae71","fff7db","ff5a47","cc0033","140200"]],
    'intensity': ["#"+tmp for tmp in ["000b14","382406","6a593c","fff7db","ff5a47","cc0033","140200"]]
}

_residuals_cranges = {
    'velocity': None,
    'linewidth': None,
    'intensity': None,
}

_attribute_colors = {
    'velocity': ["#"+tmp for tmp in ["000b14","004f8f","1f9aff","fff7db","ff5a47","cc0033","140200"]],
    'linewidth': ["#"+tmp for tmp in ["001219","005f73","0a9396","94d2bd","e9d8a6","ee9b00","ca6702","bb3e03","ae2012","9b2226"]],
    'intensity': "terrain_r",
}

_attribute_cranges = {
    'velocity': None,
    'linewidth': None,
    'intensity': "matplotlib",    
}

def use_discminer_style():
    tools = os.path.dirname(os.path.realpath(__file__))+'/tools/'
    plt.style.use(tools+'discminer.mplstyle') 

#*************
#AX DECORATORS
#*************
def mod_nticks_cbars(cbars, nbins=5):
    for cb in cbars:
        cb.locator = matplotlib.ticker.MaxNLocator(nbins=nbins)
        cb.update_ticks()
        
def mod_major_ticks(ax, axis='both', nbins=6):
    ax.locator_params(axis=axis, nbins=nbins)
    
def mod_minor_ticks(ax):
    ax.minorticks_on()
    ax.xaxis.set_minor_locator(matplotlib.ticker.AutoMinorLocator(2)) #1 minor tick per major interval
    ax.yaxis.set_minor_locator(matplotlib.ticker.AutoMinorLocator(2))

def make_up_ax(ax, xlims=(None, None), ylims=(None, None), 
               mod_minor=True, mod_major=True, **kwargs_tick_params):
    kwargs_t = dict(labeltop=True, labelbottom=False, top=True, right=True, which='both', direction='in')
    kwargs_t.update(kwargs_tick_params)
    if mod_major: mod_major_ticks(ax)
    if mod_minor: mod_minor_ticks(ax)
    ax.set_xlim(*xlims)
    ax.set_ylim(*ylims)
    ax.tick_params(**kwargs_t)

#**********************
#COLORMAP CUSTOMISATION
#**********************
def truncate_colormap(cmap, minval=0.0, maxval=1.0, n=256):
    new_cmap = matplotlib.colors.LinearSegmentedColormap.from_list(
        'trunc({n},{a:.2f},{b:.2f})'.format(n=cmap.name, a=minval, b=maxval),
        cmap(np.linspace(minval, maxval, n)))
    return new_cmap

def get_cmap_from_color(color, lev=3):
    cmap = matplotlib.colors.to_rgba(color)
    newcolors = np.tile(cmap, lev).reshape(lev,4) #Repeats the colour lev times
    newcolors[:,-1] = np.linspace(0.25, 0.95, lev) #Modifies alpha only
    new_cmap = ListedColormap(newcolors)
    return new_cmap

def mask_cmap_interval(cmap, cmap_lims, mask_lims, mask_color=np.ones(4), append=False):
    if isinstance(cmap, str): cmap = plt.get_cmap(cmap)
    cmap0, cmap1 = cmap_lims
    mask0, mask1 = mask_lims
    c0 = (mask0-cmap0)/(cmap1-cmap0)
    c1 = (mask1-cmap0)/(cmap1-cmap0)
    id0 = int(round(c0*(cmap.N)))
    id1 = int(round(c1*(cmap.N)))
    new_cmap = copy.copy(cmap)
    new_cmap._init()
    """#This block does not work, mpl does not know where to put the newly added colors
    if append:
       mask_color_arr = np.broadcast_to(mask_color, (id1-id0, 4))
       new_cmap._lut = np.insert(new_cmap._lut, id0, mask_color_arr, axis=0)
       new_cmap.N = cmap.N + id1-id0
       #Next line redoes the continuous linearsegmented colormap, thus the masked color block is reduced to a single color  
       #new_cmap = new_cmap._resample(new_cmap.N) 
    """
    new_cmap._lut[id0:id1,:] = mask_color 
    return new_cmap

def get_continuous_cmap(hex_list, float_list=None):                                                                               
    """
    Taken from https://github.com/KerryHalupka/custom_colormap 
 
    Creates and returns a color map that can be used in heat map figures.                                                             
    If float_list is not provided, the color map returned is a homogeneous gradient of the colors in hex_list.
    If float_list is provided, each color in hex_list is set to start at the corresponding location in float_list. 

    Parameters                                                                                        
    ----------                                                                                          
    hex_list: list of hex code strings                                                                
    float_list: list of floats between 0 and 1, same length as hex_list. Must start with 0 and end with 1.

    Returns     
    ----------
    matplotlib cmap

    Examples
    ----------
    fig, ax = plt.subplots(1,1)
    hex_list = ['#0091ad', '#fffffc', '#ffd166']
    x, y = np.mgrid[-5:5:0.05, -5:5:0.05]                                
    z = (np.sqrt(x**2 + y**2) + np.sin(x**2 + y**2))
    im = ax.imshow(z, cmap=get_continuous_cmap(hex_list))                                                         
    fig.colorbar(im)                                                                                                                                                                      
    ax.yaxis.set_major_locator(plt.NullLocator()) # remove y axis ticks                                                                                                                               
    ax.xaxis.set_major_locator(plt.NullLocator()) # remove x axis ticks
    plt.show()
    """

    rgb_list = [matplotlib.colors.to_rgb(i) for i in hex_list]
    if float_list is None: float_list = np.linspace(0,1,len(rgb_list))

    cdict = dict()                                                                                
    for num, col in enumerate(['red', 'green', 'blue']):                                               
        col_list = [[float_list[i], rgb_list[i][num], rgb_list[i][num]] for i in range(len(float_list))]
        cdict[col] = col_list
        cmap_new = matplotlib.colors.LinearSegmentedColormap('my_cmp', segmentdata=cdict, N=256)
    return cmap_new

def get_discminer_cmap(observable, kind='attribute'):
    if kind=='attribute':
        colors = _attribute_colors[observable]
        cranges = _attribute_cranges[observable]
    elif kind=='residuals':
        colors = _residuals_colors[observable]
        cranges = _residuals_cranges[observable]
    else:
        raise InputError(
            kind, "kind must be 'attribute' or 'residuals'"
        )

    if cranges=='matplotlib':
        cmap = copy.copy(plt.get_cmap(colors))
    elif isinstance(cranges, Iterable) or cranges is None:
        cmap = get_continuous_cmap(colors, float_list=cranges)
        
    return cmap
        
#**************
#COLORBAR STUFF
#**************
def add_cbar_ax(fig, ax, perc=4, orientation='horizontal', subplots=True):
    figx, figy = fig.get_size_inches()
    figr = figy/figx
    axp = ax.get_position()
    x0, x1, y0, y1 = axp.x0, axp.x1, axp.y0, axp.y1
    w = x1 - x0
    h = y1 - y0
    dx = dy = 0.01*perc
    
    if subplots:
        if orientation=='horizontal': return fig.add_axes([x0, y0, w, dy])
        if orientation=='vertical': return fig.add_axes([x1+0.5*dx, y0+0.1*h, dx, h-0.2*h])        
    else:
        if orientation=='horizontal': return fig.add_axes([x0, y0-0.5*dy, w, dy])
        if orientation=='vertical': return fig.add_axes([x1+0.5*dx, y0, dx, h])        
        
def make_round_cbar(ax, Rout, levels,
                    rwidth=0.06, cmap=get_discminer_cmap('velocity'),
                    quadrant=2, clabel='km/s', fmt='%4.1f'):

    sign_xy = {1: [1,1],
               2: [-1,1],
               3: [-1,-1],
               4: [1,-1],
    }[quadrant]
    
    phi0 = {1: 0.0,
            2: np.pi,
            3: -2*np.pi/3,
            4: -np.pi/3,
    }[quadrant]

    phi1 = {1: np.pi/3,
            2: 2*np.pi/3,
            3: -np.pi,
            4: 0.0,
    }[quadrant]

    rot_quad = {1: lambda phi_lev: np.degrees(phi0+phi_lev),
                2: lambda phi_lev: -np.degrees(phi0-phi_lev),
                3: lambda phi_lev: -np.degrees(phi1-phi_lev),
                4: lambda phi_lev: np.degrees(phi1+phi_lev),
    }[quadrant]

    cticks_ha = {1: 'left',
                 2: 'right',
                 3: 'right',
                 4: 'left'
    }[quadrant]

    text_R = {1: 1}
    
    phi_max = np.max([phi0,phi1])
    phi_min = np.min([phi0,phi1])
    
    xlims = ax.get_xlim()
    xcbar = np.linspace(*xlims, 1000)
    ycbar = np.linspace(-Rout, Rout, 1000)
    xx, yy = np.meshgrid(xcbar, ycbar)
    rr = np.hypot(xx, yy)
    pp = np.arctan2(yy, xx)

    cbar_phi2lev = lambda phi: (levels[-1]-levels[0])*np.abs((phi0-phi)/(phi1-phi0)) + levels[0]
    cbar_lev2phi = lambda lev: (phi1-phi0)*(lev-levels[0])/np.abs(levels[-1]-levels[0]) + phi0

    r0 = Rout+0.05*Rout
    r1 = r0+rwidth*Rout
    cbar_polar = np.zeros_like(rr) + np.nan
    mask_polar = (rr>=r0) & (rr<=r1) & (pp>=phi_min) & (pp<=phi_max)
    cbar_polar[mask_polar] = cbar_phi2lev(pp[mask_polar])
    cbar_im = ax.contourf(xx,yy, cbar_polar, levels=levels, cmap=cmap, extend='both', origin='lower') #np.ma.array(cbar_polar, mask=~mask_polar )

    cbar_levels_pol = np.linspace(levels[0], levels[-1], 5)
    cbar_levels_phi = cbar_lev2phi(cbar_levels_pol)

    for i,cbi in enumerate(cbar_levels_phi):
        Rtext = r1 + 0.03*Rout
        ax.text(Rtext*np.cos(cbi), Rtext*np.sin(cbi), fmt%cbar_levels_pol[i], fontsize=SMALL_SIZE+2, c='0.7',
                ha=cticks_ha, va='center', weight='bold', rotation_mode='anchor', rotation=rot_quad(cbi))

    ax.text(sign_xy[0]*Rtext, -sign_xy[1]*0.1*Rout, clabel, fontsize=SMALL_SIZE+3, c='0.7', ha='center', va='center', weight='bold', rotation=0)

#********************
#DISC PLOT DECORATORS
#********************
def _make_text_2D(ax, Rlist, posx=0.0, sposy=1, fmt='%d', va=None, **kwargs_text):
    if va is not None:
        _va = va
        dy = 0
    else:
        if sposy<0:
            _va = 'top'
            dy = 2
        elif sposy>0:
            _va = 'bottom'
            dy = -1
        else:
            return 0

    kwargs = dict(fontsize=SMALL_SIZE+3, ha='center', va=_va, weight='bold', rotation=0)
    kwargs.update(kwargs_text)
    for Ri in Rlist:
        ax.text(posx, sposy*(Ri+dy), fmt%Ri, **kwargs)
        
def _make_text_1D_substructures(ax, gaps=[], rings=[], kinks=[],
                                label_gaps=False, label_rings=False, label_kinks=False):
    kwargs_text = dict(fontsize=SMALL_SIZE+2, ha='center', va='bottom', transform=ax.transAxes, weight='bold', rotation=90)    
    xlims = ax.get_xlim()
    xext = xlims[1]-xlims[0]

    def text_it(R, text):
        for Ri in R:
            if Ri < xlims[0]: continue
            posx = (Ri-xlims[0])/xext
            ax.text(posx, 1.02, text%Ri, **kwargs_text)        

    if label_gaps: text_it(gaps, r'D%d')
    if label_rings: text_it(rings, r'B%d')
    if label_kinks: text_it(kinks, r'K%d')

def _make_radial_grid_2D(ax, Rout, gaps=[], rings=[], kinks=[], make_labels=True, label_freq=2):
    get_intdigits = lambda n: len(str(n).split('.')[0])
    
    angs = np.linspace(0, 2*np.pi, 100)
    cos_angs = np.cos(angs)
    sin_angs = np.sin(angs)
    
    subst = np.concatenate([gaps, rings, kinks], axis=None)
    if len(subst)>0:
        Rref = np.max(subst)
    else:
        Rref = 0.0
    Rref_digits = get_intdigits(Rref)        
        
    R_after_ref = 10**(Rref_digits-1)*ceil(Rref/(10**(Rref_digits-1)))
    Rgrid_polar = np.arange(R_after_ref, Rout, 50)
    
    for j,Ri in enumerate(Rgrid_polar[0:-1:2]):
        if Ri == Rref: continue
        ax.plot(Ri*cos_angs, Ri*sin_angs, color='k', lw=1.2, alpha=0.5,
                dash_capstyle='round', dashes=(0.5, 3.5)) 

    for j,Ri in enumerate(Rgrid_polar[1::2]):
        ax.plot(Ri*cos_angs, Ri*sin_angs, color='k', ls=':', lw=0.4, alpha=1.0)

    if make_labels:
        _make_text_2D(ax, Rgrid_polar[1::label_freq], sposy=-1, fmt='%d',
                      fontsize=SMALL_SIZE+3, color='0.1', va='center') #label in the north
        _make_text_2D(ax, Rgrid_polar[1::label_freq], sposy=1, fmt='%d',
                      fontsize=SMALL_SIZE+3, color='0.1', va='center') # and south

    ax.plot(0.98*Rout*cos_angs, 0.98*Rout*sin_angs, color='0.4', ls='-', lw=3.0, alpha=1.0)
    ax.plot(0.99*Rout*cos_angs, 0.99*Rout*sin_angs, color='0.2', ls='-', lw=3.0, alpha=1.0)
    ax.plot(1.00*Rout*cos_angs, 1.00*Rout*sin_angs, color='0.0', ls='-', lw=3.0, alpha=1.0)
    
def _make_azimuthal_grid_2D(ax, Rout):
    for j, phii in enumerate(np.arange(0, 2*np.pi, np.pi/6)):
        ax.plot([0, Rout*np.cos(phii)], [0, Rout*np.sin(phii)], color='k', ls=':', lw=0.4, alpha=1.0)

    for deg in np.linspace(0, 90, 4):
        deg_rad = np.radians(deg)
        txt = ax.text(1.04*Rout*np.cos(deg_rad), 1.04*Rout*np.sin(deg_rad), r'$%d$'%deg, c='0.0',
                      fontsize=SMALL_SIZE, ha='center', va='center', weight='bold', rotation=-(90-deg))
        txt.set_text(r'$%d^{\circ}$'%deg)

def _make_nsky_2D(ax, Rout, xlim, z_func, z_pars, incl, PA, xc=0.0, yc=0.0):
    ynorth = np.linspace(-1.1*Rout, 1.1*Rout, 100)
    xnorth = np.zeros_like(ynorth)
    xn, yn = [], []
    
    for i in range(len(ynorth)):
        xni, yni = GridTools.get_disc_from_sky_coords(xnorth[i], ynorth[i], z_func, z_pars, incl, PA, xc=0, yc=0)
        xn.append(xni)
        yn.append(yni)

    xn, yn = np.asarray(xn), np.asarray(yn)
    ax.plot(xn, yn, color='0.0', lw=1.7, dash_capstyle='round', dashes=(1.5, 2.5))
    
    text_nsky = lambda x, y: ax.text(x, y, r'$\vec{\rm N}$$_{\rm sky}$', fontsize=MEDIUM_SIZE,
                                     ha='center', va='bottom', weight='bold', rotation=np.degrees(-PA))

    if yni>xni:
        ci = yni
        cn = yn
    else:
        ci = xni
        cn = xn
        
    if ci>xlim:
        ii = np.argmin(np.abs(cn-xlim))
        text_nsky(xn[ii], yn[ii])
    else:
        text_nsky(xni, yni)

    return xni, yni

def make_substructures(ax, gaps=[], rings=[], kinks=[],
                       twodim=False, polar=False, make_legend=False,
                       label_gaps=False, label_rings=False, label_kinks=False,
                       kwargs_gaps={}, kwargs_rings={}, kwargs_kinks={}, func1d='axvline'):
    '''Overlay ring-like (if twodim) or vertical lines (if not twodim) to illustrate the radial location of substructures in the disc'''
    kwargs_g = dict(color='0.2', ls='--', lw=1.7, dash_capstyle='round', dashes=(3.0, 2.5), alpha=1.0)
    kwargs_r = dict(color='0.2', ls='-', lw=1.7, dash_capstyle='round', alpha=1.0)
    kwargs_k = dict(color='purple', ls=':', lw=2.5, dash_capstyle='round', dashes=(0.5, 1.5), alpha=0.9)
    kwargs_g.update(kwargs_gaps)
    kwargs_r.update(kwargs_rings)
    kwargs_k.update(kwargs_kinks)        
    if twodim:
        nphi = 100
        phi = np.linspace(0, 2*np.pi, nphi)
        phi_deg = np.degrees(phi-np.pi)
        subst_fmt = zip([gaps, rings, kinks], ['D%d', 'B%d', 'K%d'], [label_gaps, label_rings, label_kinks])
        if polar:
            for R in gaps: ax.plot(phi_deg, [R]*nphi, **kwargs_g)
            for R in rings: ax.plot(phi_deg, [R]*nphi, **kwargs_r)
            for R in kinks: ax.plot(phi_deg, [R]*nphi, **kwargs_k)

            for subst, fmt, label in subst_fmt:
                if label:
                    _make_text_2D(ax, subst, posx=-45, fmt=fmt)                
        else:
            cos_phi = np.cos(phi)
            sin_phi = np.sin(phi)
            for R in gaps: ax.plot(R*cos_phi, R*sin_phi, **kwargs_g)
            for R in rings: ax.plot(R*cos_phi, R*sin_phi, **kwargs_r)
            for R in kinks: ax.plot(R*cos_phi, R*sin_phi, **kwargs_k)

            for subst, fmt, label in subst_fmt:
                if label:
                    _make_text_2D(ax, subst, sposy=-1, fmt=fmt)
            
    else:
        if func1d=='axvline': func1d=ax.axvline
        elif func1d=='axhline': func1d=ax.axhline            
        for R in gaps: func1d(R, **kwargs_g)
        for R in rings: func1d(R, **kwargs_r)
        for R in kinks: func1d(R, **kwargs_k)

        _make_text_1D_substructures(ax, gaps=gaps, rings=rings, kinks=kinks,
                                    label_gaps=label_gaps, label_rings=label_rings, label_kinks=label_kinks)
        
    if make_legend and len(gaps)>0: ax.plot([None], [None], label='Gaps', **kwargs_g)
    if make_legend and len(rings)>0: ax.plot([None], [None], label='Rings', **kwargs_r)
    if make_legend and len(kinks)>0: ax.plot([None], [None], label='Kinks', **kwargs_k)

    return ax

#*********************
#MAKE DEPROJECTED MAPS
#*********************
def make_round_map(
        map2d, levels, X, Y, Rout,
        z_func=None, z_pars=None, incl=None, PA=None, xc=0, yc=0, #Optional, make N-sky axis
        fig=None, ax=None,
        rwidth=0.06, cmap=get_discminer_cmap('velocity'), clabel='km/s', fmt='%5.2f', quadrant=None, #cbar kwargs
        gaps=[], rings=[], kinks=[],
        mask_wedge=None, mask_inner=None
):

    #SOME DEFINITIONS
    if fig is None:
        fig, ax = plt.subplots(ncols=1, nrows=1, figsize=(10, 10))

    X = np.nan_to_num(X.to('au').value)
    Y = np.nan_to_num(Y.to('au').value)
    Rout = Rout.to('au').value
        
    xlim_rec = 1.15*Rout
    
    cmap_c = copy.copy(cmap)
    cmap_c.set_under('0.4')
    cmap_c.set_over('0.4')
    
    #MAIN PLOT
    im = ax.contourf(X, Y, map2d, levels=levels, cmap=cmap_c, extend='both', origin='lower')

    #RADIAL GRID
    _make_radial_grid_2D(ax, Rout, gaps=gaps, rings=rings, kinks=kinks, label_freq=2)
    
    #AZIMUTHAL GRID
    _make_azimuthal_grid_2D(ax, Rout)
    
    #SKY AXIS
    if np.all(np.asarray([z_func, z_pars, incl, PA])!=None):
        xni, yni = _make_nsky_2D(ax, Rout, xlim_rec,
                                 z_func, z_pars, incl, PA, xc=xc, yc=yc)
    else:
        xni, yni = None, None

    #SET LIMITS AND AXES
    for side in ['left','top','right']: ax.spines[side].set_visible(False)
    make_up_ax(ax,
               xlims=(-xlim_rec, xlim_rec),
               ylims=(-xlim_rec, xlim_rec),
               labelleft=False, left=False, right=False, labeltop=False, top=False, labelbottom=True, bottom=True,
               labelsize=SMALL_SIZE+3, rotation=45)
    mod_major_ticks(ax, axis='x', nbins=10)
    ax.set_xlabel('Offset [au]', fontsize=MEDIUM_SIZE+2)
    ax.set_aspect(1)

    #MAKE ROUND COLORBAR
    if quadrant is None: #Guess best quadrant based on Nsky position
        if xni is None:
            quadrant = 2
        elif (yni>0 and xni<0) or (yni<0 and xni>0):
            quadrant = 3
        else:
            quadrant = 2
            
    make_round_cbar(ax, Rout, levels, cmap=cmap_c, clabel=clabel, fmt=fmt, quadrant=quadrant)

    sq = {1: -1,
          2: 1,
          3: 1,
          4: -1,
    }[quadrant]
        
    ax.plot([sq*Rout, sq*Rout], [0, -xlim_rec], color='0.0', lw=1.0, dash_capstyle='round', dashes=(1.5, 2.5)) #Rout projected onto Cartesian xaxis

    #MASK
    kw_mask = dict(facecolor='0.5', edgecolor='k', lw=1.0, alpha=0.6)
    if mask_wedge is not None:

        for wedge in mask_wedge:

            if np.isscalar(wedge.value): #single wedge
                w = patches.Wedge((0,0), Rout, *mask_wedge.to('deg').value, **kw_mask)
                ax.add_artist(w)
                break            

            else: #list of wedges
                w = patches.Wedge((0,0), Rout, *wedge.to('deg').value, **kw_mask)
                ax.add_artist(w)
    
    if mask_inner is not None:
        inner = patches.Circle((0,0), mask_inner.to('au').value, **kw_mask)
        ax.add_artist(inner)
        
    return fig, ax


def make_polar_map(
        map2d, levels, R, PHI, Rout,
        Rin = 0.0,
        fig=None, ax=None, 
        cmap=get_discminer_cmap('velocity'),
        fmt='%5.2f', clabel=None, gradient=0,
        **kwargs_cbar
        #,filaments=[],                            
):
    from scipy.interpolate import griddata

    kwargs_cb = dict(orientation='vertical', subplots=False, perc=2)
    kwargs_cb.update(kwargs_cbar)
    
    #SOME DEFINITIONS
    if fig is None:
        fig, ax = plt.subplots(ncols=1, nrows=1, figsize=(12, 3))
    Rout = Rout.to('au').value
    try:
        Rin = Rin.to('au').value
    except AttributeError:
        Rin = Rin

    #GRID AND PLOT
    phi_nonan_deg = np.nan_to_num(PHI, nan=10*np.nanmax(PHI)).to('deg').value
    R_nonan_au = np.nan_to_num(R).to('au').value

    dR = int((Rout-Rin)/50)
    R1d = np.arange(Rin, Rout+dR, dR)
    phi1d = np.linspace(-180, 180, 1000)
    dP = phi1d[1] - phi1d[0]

    PP, RR = np.meshgrid(phi1d, R1d, indexing='xy')
    
    pmap2d = griddata((phi_nonan_deg.flatten(), R_nonan_au.flatten()), map2d.flatten(), (PP, RR), method='linear')
    
    make_plot = lambda map2d: ax.contourf(map2d, cmap=cmap,
                                          extent=[PP.min(), PP.max(), RR.min(), RR.max()],
                                          levels=levels,
                                          extend='both', origin='lower')
    
    if not gradient:
        im = make_plot(pmap2d)
        map2d = pmap2d
        
    else:
        dr, dphi = np.gradient(pmap2d) #in pix-1 units
        dr /= dR # in au-1 units
        dphi *= 1/np.radians(dP) * 1/RR # in au-1 rad-1 units 
        dmax = np.sqrt(dphi**2 + dr**2)
        
        if gradient=='phi':
            im = make_plot(dphi)
            map2d = dphi
        elif gradient=='r':
            im = make_plot(dr)
            map2d = dr
        elif gradient=='peak':
            im = make_plot(dmax)
            map2d = dmax
        else:
            raise InputError(
                kind, "kind must be 'attribute' or 'residuals'"
            )
   
        
    """
    kw_fil = dict(s=20, lw=1, marker='+')
    for i,filament in enumerate(filaments):
        if i==0 and filament is not None:
            kw_fil.update(dict(fc='red', ec='firebrick'))
        elif i==1 and filament is not None:
            kw_fil.update(dict(fc='blue', ec='navy'))
        rrr=R_nonan_up_au[filament.skeleton.astype('bool')]
        ppr=phi_nonan_up_deg[filament.skeleton.astype('bool')]
        ax.scatter(ppr, rrr, **kw_fil)
    """

    #DECORATIONS
    ax.axvline(-90, ls=':', lw=2.5, color='0.3', dash_capstyle='round')
    ax.axvline(90, ls=':', lw=2.5, color='0.3', dash_capstyle='round')
    ax.axvline(0, ls=':', lw=2.5, color='0.3', dash_capstyle='round')

    make_up_ax(ax, labelbottom=True, labeltop=False, labelsize=SMALL_SIZE+1,
               xlims=(-180.1, 180.1), ylims=(1.1*Rin, 0.95*Rout))
    ax.set_xticks(np.linspace(-180, 180, 13))
    ax.set_xlabel('Azimuth [deg]', fontsize=MEDIUM_SIZE)
    ax.set_ylabel('Radius [au]', fontsize=MEDIUM_SIZE)
    mod_major_ticks(ax, axis='y', nbins=10)
                       
    cax = add_cbar_ax(fig, ax, **kwargs_cb)
    cbar = plt.colorbar(im, cax=cax, format=fmt, orientation='vertical', ticks=np.linspace(levels.min(), levels.max(), 5))
    cbar.ax.tick_params(which='major', direction='in', width=2.7, size=4.8, pad=4, labelsize=SMALL_SIZE)
    cbar.ax.tick_params(which='minor', direction='in', width=2.7, size=3.3)
    cbar.set_label(clabel, fontsize=SMALL_SIZE, labelpad=20, rotation=270)
    mod_minor_ticks(cbar.ax)

    return fig, ax, cbar


def make_pie_map(
        X, Y, Rout,
        quadrant_map2d = {1: None, 2: None, 3: None, 4: None},
        quadrant_levels = {1: None, 2: None, 3: None, 4: None},
        quadrant_cmap = {1: 'gnuplot2', 2: 'jet', 3: 'plasma', 4: 'seismic'},
        quadrant_clabel = {1: None, 2: None, 3: None, 4: None},
        quadrant_fmt = {1: '%4d', 2: '%4d', 3: '%4d', 4: '%4d'},        
        gaps=[], rings=[], kinks=[],
        fig=None, ax=None,        
):
    #SOME DEFINITIONS
    if fig is None:
        fig, ax = plt.subplots(ncols=1, nrows=1, figsize=(10, 10))

    Rout = Rout.to('au').value
    X = X.to('au').value
    Y = Y.to('au').value

    quadrant_reg = {4: ((X>=0) & (Y<0)),
                    3: ((X<0) & (Y<=0)),
                    2: ((X<=0) & (Y>0)),
                    1: ((X>0) & (Y>=0))
    }
    
    #CBAR STUFF
    figx, figy = fig.get_size_inches()
    figr = figy/figx
    axp = ax.get_position()
    x0, x1, y0, y1 = axp.x0, axp.x1, axp.y0, axp.y1
    w = x1 - x0
    h = y1 - y0
    dx = 0.02
    dy = 6*dx

    cbar_pos = {1: [x1, y1-0.4*h, dx, dy],
                2: [x0-dx, y1-0.4*h, dx, dy],
                3: [x0-dx, y0+0.4*h-dy, dx, dy],
                4: [x1, y0+0.4*h-dy, dx, dy]}

    #MAKE ALL
    for quad in [1,2,3,4]:
        inquad = quadrant_reg[quad]
        map2d = quadrant_map2d[quad]
        levels = quadrant_levels[quad]
        cmap = quadrant_cmap[quad]
        clabel = quadrant_clabel[quad]
        cfmt = quadrant_fmt[quad]
        
        if map2d is None:
            map2d = np.zeros_like(X)
        else:
            map2d[~inquad] = np.nan

        if levels is None:
            levels = np.linspace(np.nanmin(map2d), np.nanmax(map2d), 64)
            
        im = ax.contourf(X, Y, map2d, levels=levels, extend='both', cmap=cmap)
    
        ax_cbar = fig.add_axes(cbar_pos[quad])
        cbar_ticks = np.linspace(np.min(levels), np.max(levels), 5)
        cbar = plt.colorbar(im, cax=ax_cbar, format=cfmt, orientation='vertical', ticks=cbar_ticks)
        cbar.ax.tick_params(which='major', direction='in', width=2.7, size=4.8, pad=7, labelsize=MEDIUM_SIZE-2)
        cbar.ax.tick_params(which='minor', direction='in', width=2.7, size=3.3)
        cbar.set_label(clabel, fontsize=MEDIUM_SIZE, rotation=90, labelpad=7)        
        mod_minor_ticks(cbar.ax)

        if quad==2 or quad==3:
            cbar.ax.tick_params(which='both', left=True, right=False, labelleft=True, labelright=False)
        else:
            cbar.ax.yaxis.set_label_position('left')
            
    _make_radial_grid_2D(ax, Rout, gaps=gaps, rings=rings, kinks=kinks, make_labels=True, label_freq=2)

    for side in ['left','top','right']: ax.spines[side].set_visible(False)
    for side in ['left']: ax.spines[side].set_linewidth(3.5)

    make_up_ax(ax,
               xlims=(-1.15*Rout, 1.15*Rout),
               ylims=(-Rout-5, Rout+5),
               labelleft=False, left=False, right=False,
               labeltop=False, top=False,
               labelbottom=True, bottom=True,
               labelsize=MEDIUM_SIZE+2, rotation=45)
    mod_major_ticks(ax, axis='x', nbins=10)
    
    ax.set_xlabel('Offset [au]', fontsize=MEDIUM_SIZE+2)
    ax.set_aspect(1)
    
    return fig, ax


#***************************
#PEAK RESIDUALS AND CLUSTERS
#***************************
def append_sigma_panel(fig, ax, values, ax_std=None, weights=None, hist=False, fit_gauss_hist=False): #attach sigma panel to AxesSubplot, based on input values
    ax = np.atleast_1d(ax)

    if ax_std is None:
        axp = ax[-1].get_position()
        ax = np.append(ax, fig.add_axes([axp.x0 + axp.width, axp.y0, 0.1, axp.height]))
    else:
        ax = np.append(ax, ax_std)
        
    #gauss = lambda x, A, mu, sigma: A*np.exp(-(x-mu)**2/(2.*sigma**2))
    ax1_ylims = ax[-2].get_ylim()

    for axi in ax[:-1]:
        axi.tick_params(which='both', right=False, labelright=False)

    ax[-1].tick_params(which='both', top=False, bottom=False, labelbottom=False, 
                       left=False, labelleft=False, right=True, labelright=True)
    ax[-1].yaxis.set_label_position('right')
    ax[-1].spines['left'].set_color('0.6')
    ax[-1].spines['left'].set_linewidth(3.5)

    if weights is not None:
        values_mean = np.sum(weights*values)/np.sum(weights)
        values_std = weighted_std(values, weights, weighted_mean=values_mean)
    else:
        values_mean = np.mean(values)
        values_std = np.std(values)

    max_y = 1.0

    if hist:
        n, bins, patches = ax[-1].hist(values, bins=2*int(round(len(values)**(1/3.)))-1, orientation='horizontal', 
                                       density=True, linewidth=1.5, facecolor='0.95', edgecolor='k', alpha=1.0)
        max_y = np.max(n)
        if fit_gauss_hist: #Fit Gaussian to histogram to compare against data distribution
            coeff, var_matrix = curve_fit(_gauss, 0.5*(bins[1:]+bins[:-1]), n, p0=[max_y, values_mean, values_std])
            values_x = np.linspace(values_mean-4*values_std, values_mean+4*values_std, 100)
            values_y = _gauss(values_x, *coeff)
            ax[-1].plot(values_y, values_x, color='tomato', ls='--', lw=2.0)

    values_x = np.linspace(values_mean-4*values_std, values_mean+4*values_std, 100)
    values_pars =  [max_y, values_mean, values_std]
    values_y = _gauss(values_x, *values_pars)
    ax[-1].plot(values_y, values_x, color='limegreen', lw=3.5)
    ax[-1].set_xlim(-0.2*max_y, 1.2*max_y)

    for i in range(0,4): 
        values_stdi = values_mean+i*values_std
        gauss_values_stdi = _gauss(values_stdi, *values_pars)
        ax[-1].plot([-0.2*max_y, gauss_values_stdi], [values_stdi]*2, color='0.6', ls=':', lw=2.)
        for axi in ax[:-1]: axi.axhline(values_stdi, color='0.6', ls=':', lw=2.)
        if values_stdi < ax1_ylims[-1] and i>0:
            ax[-1].text(gauss_values_stdi+0.2*max_y, values_stdi, r'%d$\sigma$'%i, 
                        fontsize=14, ha='center', va='center', rotation=-90)

    for axi in ax:
        axi.set_ylim(*ax1_ylims)




from .utils import FrontendUtils, InputError, FITSUtils, get_tb

import numpy as np
from astropy.io import fits
from scipy.optimize import curve_fit

import copy

_progress_bar = FrontendUtils._progress_bar

def _gauss(x, *p):
    A1, mu1, sigma1 = p
    return A1*np.exp(-(x-mu1)**2/(2.*sigma1**2))

def _doublegauss_sum(x, *p):
    A1, mu1, sigma1, A2, mu2, sigma2 = p
    return A1*np.exp(-(x-mu1)**2/(2.*sigma1**2)) + A2*np.exp(-(x-mu2)**2/(2.*sigma2**2))

def _doublegauss_mask(x, *p):
    A1, mu1, sigma1, A2, mu2, sigma2 = p
    gauss1 = A1*np.exp(-(x-mu1)**2/(2.*sigma1**2))
    gauss2 = A2*np.exp(-(x-mu2)**2/(2.*sigma2**2))
    return np.where(gauss1>=gauss2, gauss1, gauss2)

def _bell(x, *p):
    A1, mu1, sigma1, Ls1 = p
    return A1/(1+np.abs((x-mu1)/sigma1)**(2*Ls1))

def _doublebell_sum(x, *p):
    A1, mu1, sigma1, Ls1, A2, mu2, sigma2, Ls2 = p
    return A1/(1+np.abs((x-mu1)/sigma1)**(2*Ls1)) + A2/(1+np.abs((x-mu2)/sigma2)**(2*Ls2))

def _doublebell_mask(x, *p):
    A1, mu1, sigma1, Ls1, A2, mu2, sigma2, Ls2 = p
    bell1 = A1/(1+np.abs((x-mu1)/sigma1)**(2*Ls1))
    bell2 = A2/(1+np.abs((x-mu2)/sigma2)**(2*Ls2))
    return np.where(bell1>=bell2, bell1, bell2)

def _not_available(method):
    raise InputError(method, "requested moment method is currently unavailable")

def get_channels_from_parcube(parcube_up, parcube_low, vchannels, method='doublebell', kind='mask', n_fit=None):
    
    nx, ny = parcube_up.shape[1:]
    intensity = np.zeros((len(vchannels), ny, nx))
    
    if parcube_low is None and 'double' in method:
        parcube_low = np.zeros_like(parcube_up)

    par_double = lambda i,j: np.append(parcube_up[:,i,j], parcube_low[:,i,j])
    par_single = lambda i,j: parcube_up[:,i,j]
    
    if 'double' in method:
        pars = par_double
    else:
        pars = par_single

    #******************************
    #KERNELS AND RELEVANT FUNCTIONS
    if method=='doublegaussian':
        fit_func1d = _gauss
        if kind=='sum':
            fit_func = _doublegauss_sum
        elif kind=='mask':
            fit_func = _doublegauss_mask
        else:
            raise InputError(kind, "kind must be 'mask' or 'sum'")

    elif method=='doublebell':
        fit_func1d = _bell
        if kind=='sum':
            fit_func = _doublebell_sum
        elif kind=='mask':
            fit_func = _doublebell_mask
        else:
            raise InputError(kind, "kind must be 'mask' or 'sum'")

    elif method=='gaussian':
        fit_func = _gauss

    elif method=='bell':
        fit_func = _bell
        
    else:
        _not_available(method)

    #***********************
    #EVALUATE PARS ON KERNEL
    for i in range(ny):
        for j in range(nx):
            intensity[:,i,j] = fit_func(vchannels, *pars(i,j))
    
    return intensity


def fit_twocomponent(cube, model=None, lw_chans=1.0, lower2upper=1.0,
                     method='doublegaussian', kind='mask', sigma_thres=5,
                     sigma_fit=None,
):

    data = cube.data    
    vchannels = cube.vchannels
    dv = np.mean(vchannels[1:]-vchannels[:-1])
    lw_sign = np.sign(dv)
    
    nchan, nx, ny = np.shape(data)
    n_one, n_two, n_bad, n_hot, n_mask = 0, 0, 0, 0, 0
    n_fit = np.zeros((nx, ny))
    
    peak_up, dpeak_up = np.zeros((2, nx, ny))
    centroid_up, dcentroid_up = np.zeros((2, nx, ny))
    linewidth_up, dlinewidth_up = np.zeros((2, nx, ny))
    lineslope_up, dlineslope_up = np.zeros((2, nx, ny))    
    peak_low, dpeak_low = np.zeros((2, nx, ny))
    centroid_low, dcentroid_low = np.zeros((2, nx, ny))
    linewidth_low, dlinewidth_low = np.zeros((2, nx, ny))
    lineslope_low, dlineslope_low = np.zeros((2, nx, ny))
    
    is_dbell = method=='doublebell'
    
    #MODEL AS INITIAL GUESS?
    if model is None:
        print ('Guessing upper surface properties from data to use them as seeds for both upper (primary) and lower (secondary) surface components ...')        
        ind_max = np.nanargmax(data, axis=0)
        I_max = np.nanmax(data, axis=0)
        vel_peak = vchannels[ind_max]
        I_max_upper = I_max
        I_max_lower = lower2upper*I_max
        vel_peak_upper = vel_peak_lower = vel_peak
        lw_upper = lw_lower = lw_chans*dv*np.ones_like(vel_peak)
        ls_upper = ls_lower = 1.5
        
    else:
        vel2d, int2d, linew2d, lineb2d = model.props
        R, phi, z = [model.projected_coords[key] for key in ['R', 'phi', 'z']]
        I_upper = int2d['upper']
        I_lower = int2d['lower']
        print ('Using upper and lower surface properties from discminer model as initial guesses...')
        
        if np.any(np.array(['K', 'Kelvin', 'K ', 'Kelvin ']) == cube.header['BUNIT']): #If input unit is K the raw model intensity must be converted    
            r"""
            mheader = copy.copy(cube.header)
            mheader["BUNIT"] = "beam-1 Jy"
            I_upper = FITSUtils._convert_to_tb(I_upper*cube.beam_area, mheader, cube.beam, planck=False, writefits=False)
            #"""
            restfreq = cube.header['RESTFRQ']*1e-9
            I_upper = get_tb(1e3*I_upper*cube.beam_area, restfreq, cube.beam, full=False) #I in mJy/beam
            I_lower = get_tb(1e3*I_lower*cube.beam_area, restfreq, cube.beam, full=False) 

        #Initial guesses
        ind_max = np.nanargmax(data, axis=0)
        cube_max = np.take_along_axis(data, ind_max[None], axis=0).squeeze()
        I_max_upper = np.where(np.isnan(I_upper), 1.0*cube_max, I_upper)
        I_max_lower = np.where(np.isnan(I_lower), 0.5*cube_max, I_lower)
        vel_peak_upper = np.where(np.isnan(vel2d['upper']), vchannels[ind_max], vel2d['upper'])
        vel_peak_lower = np.where(np.isnan(vel2d['lower']), vchannels[ind_max], vel2d['lower'])
        lw_upper = lw_sign*np.where(np.isnan(linew2d['upper']), lw_sign*1.5*dv, linew2d['upper'])
        lw_lower = lw_sign*np.where(np.isnan(linew2d['lower']), lw_sign*1.5*dv, linew2d['lower'])
        ls_upper = np.where(np.isnan(linew2d['upper']), 1.5, lineb2d['upper'])
        ls_lower = np.where(np.isnan(linew2d['lower']), 1.5, lineb2d['lower'])

    if sigma_fit is None: sigma_func = lambda i,j: None
    else: sigma_func = lambda i,j: sigma_fit[:,i,j]

    noise = np.std( np.append(data[:5,:,:], data[-5:,:,:], axis=0), axis=0) #rms intensity from first and last 5 channels
    mask = np.nanmax(data, axis=0) <= sigma_thres*noise

    #******************************
    #KERNELS AND RELEVANT FUNCTIONS
    if method=='doublegaussian':
        fit_func1d = _gauss
        pfunc_two = lambda i,j: [I_max_upper[i,j], vel_peak_upper[i,j], lw_upper[i,j],
                                 I_max_lower[i,j], vel_peak_lower[i,j], lw_lower[i,j]]
        pfunc_one = lambda i,j: [I_max_upper[i,j], vel_peak_upper[i,j], lw_upper[i,j]]
        if kind=='sum':
            fit_func = _doublegauss_sum
        elif kind=='mask':
            fit_func = _doublegauss_mask
        else:
            raise InputError(kind, "kind must be 'mask' or 'sum'")
        idlow = np.array([3,4,5])
        bound0 = (0, -np.inf, -np.inf, 0, -np.inf, -np.inf)
        bound1 = (np.inf, np.inf, np.inf, np.inf, np.inf, np.inf)

    elif method=='doublebell':
        fit_func1d = _bell
        pfunc_two = lambda i,j: [I_max_upper[i,j], vel_peak_upper[i,j], lw_upper[i,j], ls_upper[i,j],
                                 I_max_lower[i,j], vel_peak_lower[i,j], lw_lower[i,j], ls_lower[i,j]]
        pfunc_one = lambda i,j: [I_max_upper[i,j], vel_peak_upper[i,j], lw_upper[i,j], ls_upper[i,j]]
        
        if kind=='sum':
            fit_func = _doublebell_sum
        elif kind=='mask':
            fit_func = _doublebell_mask
        else:
            raise InputError(kind, "kind must be 'mask' or 'sum'")
        idlow = np.array([4,5,6,7])
        bound0 = (0, -np.inf, -np.inf, 0, 0, -np.inf, -np.inf, 0)
        bound1 = (np.inf, np.inf, np.inf, np.inf, np.inf, np.inf, np.inf, np.inf)

    else:
        _not_available(method)

    print ('Fitting two-component function along velocity axis of the input cube...')

    #********
    #MAKE FIT
    for i in range(nx):
        for j in range(ny):
            tmp_data = data[:,i,j]

            if mask[i,j]:
                n_mask += 1
                n_fit[i,j] = -10                                    
                continue
            
            try: 
                coeff, var_matrix = curve_fit(fit_func,
                                              vchannels, tmp_data,
                                              p0=pfunc_two(i,j),
                                              ftol=1e-10, xtol=1e-10, gtol=1e-10, method='lm')
                                              #bounds = [bound0, bound1])
                deltas = np.sqrt(np.abs(np.diag(var_matrix)))
                n_two += 1
                n_fit[i,j] = 2

            except RuntimeError:
                try: 
                    coeff, var_matrix = curve_fit(fit_func1d,
                                                  vchannels, tmp_data,
                                                  p0=pfunc_one(i,j)
                    )
                    coeff = np.append(coeff, coeff)
                    deltas = np.sqrt(np.abs(np.diag(var_matrix)))
                    deltas = np.append(deltas, deltas)
                    n_one += 1
                    n_fit[i,j] = 1
                    
                except RuntimeError:
                    n_bad += 1
                    n_fit[i,j] = 0                    
                    continue

            r"""
            #
            if np.abs(coeff[idlow[1]]-vel_peak_upper[i,j]) < np.abs(coeff[1]-vel_peak_upper[i,j]):
                coeff = np.append(coeff[idlow[0]:idlow[-1]+1], coeff[0:idlow[0]])
                deltas = np.append(deltas[idlow[0]:idlow[-1]+1], deltas[0:idlow[0]])
            #"""
            r"""
            #Sometimes the fit can output upper profiles near to zero intensity and in turn the lower profile is attributed to the bulk of the emission.
            #The following seems to improve things due to this in a few pixels. However, the conditional above accounts for this for the most part already.
            if coeff[0]<=1.0 and coeff[3]>1.0:
                coeff = np.append(coeff[idlow[0]:idlow[-1]+1], coeff[0:idlow[0]])
                deltas = np.append(deltas[idlow[0]:idlow[-1]+1], deltas[0:idlow[0]])
            #""" 
            peak_up[i,j] = coeff[0]
            centroid_up[i,j] = coeff[1]
            linewidth_up[i,j] = coeff[2]
            
            peak_low[i,j] = coeff[idlow[0]]
            centroid_low[i,j] = coeff[idlow[1]]
            linewidth_low[i,j] = coeff[idlow[2]]

            dpeak_up[i,j] = deltas[0]
            dcentroid_up[i,j] = deltas[1]
            dlinewidth_up[i,j] = deltas[2]

            dpeak_low[i,j] = deltas[idlow[0]]
            dcentroid_low[i,j] = deltas[idlow[1]]
            dlinewidth_low[i,j] = deltas[idlow[2]]

            if is_dbell:
                lineslope_up[i,j] = coeff[3]
                lineslope_low[i,j] = coeff[idlow[3]]
                dlineslope_up[i,j] = deltas[3]
                dlineslope_low[i,j] = deltas[idlow[3]]
                        
        _progress_bar(int(100*i/nx))
    _progress_bar(100)

    #*****************************
    #KEEPING TRACK OF 'HOT' PIXELS
    #*****************************
    #Hot pixels will be tagged as -1
    mm = n_fit == -10 #noise
    ii = ((peak_up < 0.0) | (peak_low < 0.0)) & (~mm) #negative intensities
    peak_thres = 2*np.nanmax(cube_max)
    jj = ((peak_up > peak_thres) | (peak_low > peak_thres)) & (~mm) #too large intensities
    cc = (centroid_up == centroid_low) & (n_fit != 1) & (~mm) #up==low
    ww = ((np.abs(linewidth_up) <= 0.5*np.abs(dv)) | (np.abs(linewidth_low) <= 0.5*np.abs(dv))) & (~mm) #narrow component
    n_fit[ii+jj+cc+ww] = -1
    n_hot = np.sum(ii+cc+ww)
    
    #***************
    #PACK AND RETURN
    #***************    
    upper = [peak_up, centroid_up, linewidth_up]
    dupper = [dpeak_up, dcentroid_up, dlinewidth_up]
    lower = [peak_low, centroid_low, linewidth_low]
    dlower = [dpeak_low, dcentroid_low, dlinewidth_low]

    if is_dbell:
        upper += [lineslope_up]
        dupper += [dlineslope_up]
        lower += [lineslope_low]
        dlower += [dlineslope_low]
        
    print ('\nTwo-component fit did not converge for %.2f%s of the pixels'%(100.0*(n_bad)/(nx*ny),'%'))
    print ('A single component was fit for %.2f%s of the pixels'%(100.0*(n_one)/(nx*ny),'%'))
    print ('Masked pixels below intensity threshold: %.2f%s'%(100.0*(n_mask)/(nx*ny),'%'))
    print ('Hot pixels: %.2f%s'%(100.0*(n_hot)/(nx*ny),'%'))        
    
    return upper, dupper, lower, dlower, n_fit


def fit_gaussian(*args, **kwargs): #Backcompat
    return fit_onecomponent(*args, **kwargs)

def fit_onecomponent(
        cube, method='gaussian', lw_chans=1.0, peak_kernel=True, sigma_fit=None, sigma_thres=4
):
    """
    Fit 'gaussian' or 'bell' profiles along the velocity axis of the input cube.
    
    Parameters
    ----------
    lw_chans : int, optional
        Number of channel widths to take as initial guess for line width

    sigma_fit : array_like, optional 
        2-D array of weights computed for each pixel. Shape must be equal to that of the spatial axes of the input data. 

    peak_kernel : bool, optional
        If True (default) the returned amplitude is the peak of the kernel fitted to the line. Otherwise, the actual peak of the line profile is returned.

    Returns
    -------
    A,B,C,dA,dB,dC : 2-D arrays
        Amplitude, centre, standard deviation, and their corresponding errors, obtained from Gaussians fitted to the input datacube along its velocity axis, on each spatial pixel.
    
    """    
    data = cube.data
    vchannels = cube.vchannels

    nchan, nx, ny = np.shape(data)
    n_fit = np.zeros((nx, ny))                        
    n_bad, n_mask = 0, 0
    
    peak, dpeak = np.zeros((nx, ny)), np.zeros((nx, ny))
    centroid, dcent = np.zeros((nx, ny)), np.zeros((nx, ny))
    linewidth, dlinew = np.zeros((nx, ny)), np.zeros((nx, ny))
    lineslope, dlineslope = np.zeros((nx, ny)), np.zeros((nx, ny))    

    is_bell = method=='bell'
    
    ind_max = np.nanargmax(data, axis=0)
    I_max = np.nanmax(data, axis=0)
    vel_peak = vchannels[ind_max]
    dv = lw_chans*np.mean(vchannels[1:]-vchannels[:-1])
    Ls = 2.0 #p0 Line slope
    
    if sigma_fit is None: sigma_func = lambda i,j: None
    else: sigma_func = lambda i,j: sigma_fit[:,i,j]

    #******************************
    #KERNELS AND RELEVANT FUNCTIONS
    if method in ['gaussian', 'gauss']:
        fit_func1d = _gauss
        pfunc_one = lambda i,j: [I_max[i,j], vel_peak[i,j], dv]
        bound0 = (0, -np.inf, -np.inf, 0, -np.inf, -np.inf)
        bound1 = (np.inf, np.inf, np.inf, np.inf, np.inf, np.inf)

    elif method=='bell':
        fit_func1d = _bell
        pfunc_one = lambda i,j: [I_max[i,j], vel_peak[i,j], dv, Ls]
        bound0 = (0, -np.inf, -np.inf, 0, 0, -np.inf, -np.inf, 0)
        bound1 = (np.inf, np.inf, np.inf, np.inf, np.inf, np.inf, np.inf, np.inf)
    else:
        _not_available(method)

    noise = np.std( np.append(data[:5,:,:], data[-5:,:,:], axis=0), axis=0) #rms intensity from first and last 5 channels
    mask = np.nanmax(data, axis=0) <= sigma_thres*noise
        
    print ('Fitting one-component function along velocity axis of the input cube...')

    for i in range(nx):
        for j in range(ny):
            tmp_data = data[:,i,j]
            
            if mask[i,j]:
                n_mask += 1
                n_fit[i,j] = -10                                    
                continue
            
            try:
                coeff, var_matrix = curve_fit(fit_func1d, vchannels, tmp_data, p0=pfunc_one(i,j), sigma=sigma_func(i,j))
                n_fit[i,j] = 1
                    
            except RuntimeError: 
                n_bad+=1
                n_fit[i,j] = 0                
                continue

            deltas = np.sqrt(np.diag(var_matrix))
            
            if peak_kernel: peak[i,j] = coeff[0]
            else: peak[i,j] = I_max[i,j]
            centroid[i,j] = coeff[1]
            linewidth[i,j] = coeff[2]
            dpeak[i,j], dcent[i,j], dlinew[i,j] = deltas[:3]

            if is_bell:
                lineslope[i,j] = coeff[3]
                dlineslope[i,j] = deltas[3]
                
        _progress_bar(int(100*i/nx))

    _progress_bar(100)

    #***************
    #PACK AND RETURN
    #***************    
    upper = [peak, centroid, np.abs(linewidth)]
    dupper = [dpeak, dcent, dlinew]

    if is_bell:
        upper += [lineslope]
        dupper += [dlineslope]
    
    print ('\nOne-component fit did not converge for %.2f%s of the pixels'%(100.0*n_bad/(nx*ny),'%'))
    print ('Masked pixels below intensity threshold: %.2f%s'%(100.0*(n_mask)/(nx*ny),'%'))
    
    return upper, dupper, n_fit



from discminer.core import Data
from discminer.rail import Contours
import discminer.cart as cart
from discminer.plottools import (make_round_map,
                                 make_polar_map,
                                 make_substructures,
                                 make_up_ax,
                                 mod_major_ticks,
                                 mod_nticks_cbars,
                                 use_discminer_style)

from utils import (get_2d_plot_decorators,
                   get_noise_mask,
                   load_moments,
                   load_disc_grid,
                   add_parser_args)

import numpy as np
import matplotlib.pyplot as plt
from astropy import units as u

import json
from argparse import ArgumentParser

use_discminer_style()

parser = ArgumentParser(prog='plot residual maps', description='Plot residual maps')
args = add_parser_args(parser, moment=True, kind=True, surface=True, projection=True)

#**********************
#JSON AND PARSER STUFF
#**********************
with open('parfile.json') as json_file:
    pars = json.load(json_file)

meta = pars['metadata']
best = pars['best_fit']
custom = pars['custom']

vsys = best['velocity']['vsys']
Rout = best['intensity']['Rout']
incl = best['orientation']['incl']
PA = best['orientation']['PA']
xc = best['orientation']['xc']
yc = best['orientation']['yc']

gaps = custom['gaps']
rings = custom['rings']

ctitle, clabel, clim, cfmt, cmap_mom, cmap_res, levels_im, levels_cc, unit = get_2d_plot_decorators(args.moment, unit_simple=True, fmt_vertical=True)

#****************
#SOME DEFINITIONS
#****************
file_data = meta['file_data']
tag = meta['tag']
au_to_m = u.au.to('m')

dpc = meta['dpc']*u.pc
Rmax = 1.1*Rout*u.au #Max model radius, 10% larger than disc Rout

#********************
#LOAD DATA AND GRID
#********************
datacube = Data(file_data, dpc) # Read data and convert to Cube object
noise_mean, mask = get_noise_mask(datacube, thres=2)

#Useful definitions for plots
with open('grid_extent.json') as json_file:
    grid = json.load(json_file)

xmax = grid['xsky'] 
xlim = 1.15*np.min([xmax, Rmax.value])
extent= np.array([-xmax, xmax, -xmax, xmax])

#*************************
#LOAD DISC GEOMETRY
R, phi, z = load_disc_grid()

Xproj = R[args.surface]*np.cos(phi[args.surface])
Yproj = R[args.surface]*np.sin(phi[args.surface])

#*************************
#LOAD MOMENT MAPS    
moment_data, moment_model, mtags = load_moments(args)

#****************
#USEFUL FUNCTIONS
#****************
def clip_prop_radially(prop2d, Rmin=datacube.beam_size, Rmax=np.inf, Rgrid=R[args.surface]*u.m):
    Rmin = Rmin.to('au').value
    Rgrid = np.nan_to_num(Rgrid).to('au').value
    try:
        Rmax = Rmax.to('au').value
    except AttributeError:
        Rmax = Rmax
    return np.where((Rgrid<Rmin) | (Rgrid>Rmax), np.nan, prop2d)

#**************************
#MASK AND COMPUTE RESIDUALS
moment_data = np.where(mask, np.nan, moment_data)
moment_model = np.where(mask, np.nan, moment_model)
residuals = clip_prop_radially(moment_data - moment_model, Rmax=Rout)

clabels = {
    'linewidth': r'$\Delta$ Line width [km s$^{-1}$]',
    'lineslope': r'$\Delta$ Line slope',
    'velocity': r'$\Delta$ Centroid [km s$^{-1}$]',
    'peakintensity': r'$\Delta$ Peak Int. [K]'
}

if args.projection=='cartesian':
    levels_resid = np.linspace(-clim, clim, 32)
    
    if args.surface in ['up', 'upper']:
        z_func = cart.z_upper_exp_tapered
        z_pars = best['height_upper']

    elif args.surface in ['low', 'lower']:
        z_func = cart.z_lower_exp_tapered
        z_pars = best['height_lower']
    
    fig, ax = make_round_map(residuals, levels_resid, Xproj*u.m, Yproj*u.m, Rout*u.au,
                             z_func=z_func, z_pars=z_pars, incl=incl, PA=PA, xc=xc, yc=yc,
                             cmap=cmap_res, clabel=unit, fmt=cfmt, 
                             gaps=gaps, rings=rings)
    
    make_substructures(ax, gaps=gaps, rings=rings, twodim=True, label_rings=True)

elif args.projection=='polar':
    levels_resid = np.linspace(-clim, clim, 48)    
    fig, ax, cbar = make_polar_map(residuals, levels_resid,
                                   R[args.surface]*u.m, phi[args.surface]*u.rad, Rout*u.au,
                                   Rin = datacube.beam_size,
                                   cmap=cmap_res, fmt=cfmt, clabel=clabels[args.moment])
                                   
    make_substructures(ax, gaps=gaps, rings=rings, twodim=True, polar=True, label_rings=True)
    
ax.set_title(ctitle, fontsize=16, color='k')

plt.savefig('residuals_deproj_%s_%s.png'%(mtags['base'], args.projection), bbox_inches='tight', dpi=200)
plt.show()
plt.close()



"""A setuptools based setup module.

See:
https://packaging.python.org/guides/distributing-packages-using-setuptools/
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
import pathlib
import re

# Get version
VERSIONFILE="discminer/_version.py"
verstrline = open(VERSIONFILE, "rt").read()
VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(VSRE, verstrline, re.M)
if mo:
    verstr = mo.group(1)
else:
    raise RuntimeError("Unable to find version string in %s." % (VERSIONFILE,))
#*************

here = pathlib.Path(__file__).parent.resolve()

# Get the long description from the README file
long_description = (here / 'README.md').read_text(encoding='utf-8')

setup(
    # $ pip install discminer
    # https://pypi.org/project/discminer/
    name='discminer',  # Required
    # https://packaging.python.org/guides/single-sourcing-package-version/
    version=verstr,  # Required
    # https://packaging.python.org/specifications/core-metadata/#summary
    description='Python package for parametric modelling of intensity channel maps from gas discs',  # Optional
    # https://packaging.python.org/specifications/core-metadata/#description-optional
    long_description=long_description,  # Optional
    # https://packaging.python.org/specifications/core-metadata/#description-content-type-optional
    long_description_content_type='text/markdown',  # Optional (see note above)
    # https://packaging.python.org/specifications/core-metadata/#home-page-optional
    url='https://github.com/andizq/discminer',  # Optional
    author='Andres F. Izquierdo',  # Optional
    author_email='andres.izquierdo.c@gmail.com',  # Optional
    # For a list of valid classifiers, see https://pypi.org/classifiers/
    classifiers=[  # Optional
        'Development Status :: 4 - Beta',
        'Topic :: Scientific/Engineering :: Astronomy',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3 :: Only',
    ],
    keywords='astronomy, discs, disks, planets, detection',  # Optional
    #package_dir={'discminer': 'discminer'},  # Optional
    packages=['discminer', 'discminer.tools'],  # Required
    # https://packaging.python.org/guides/distributing-packages-using-setuptools/#python-requires
    python_requires='>=3.6, <4',
    # https://packaging.python.org/discussions/install-requires-vs-requirements/
    install_requires=
    [
        'numpy>=1.18',
        'scipy>=1.5',
        'matplotlib>=3',
        'astropy>=3',
        'emcee>=3',
        'radio-beam',
        'scikit-image',
        'scikit-learn',
        'spectral-cube'
    ], 
    # If there are data files included in your packages that need to be
    # installed, specify them here.
    package_data={  # Optional
        'discminer':
        [
            'icons/logo.txt',
            'icons/button*',
            'tools/discminer.mplstyle'
        ],
    },
    # https://packaging.python.org/specifications/core-metadata/#project-url-multiple-use
    project_urls={  # Optional
        'Bug Reports': 'https://github.com/andizq/discminer/issues',
        'Source': 'https://github.com/andizq/discminer/',
    },
)



from discminer.plottools import (get_discminer_cmap, append_sigma_panel,
                                 make_up_ax, mod_minor_ticks, mod_major_ticks,
                                 use_discminer_style, mod_nticks_cbars,
                                 make_substructures, make_round_map)
from discminer.pick import Pick
from discminer.rail import Contours

import numpy as np
import matplotlib.pyplot as plt

from astropy import units as u
import json
import copy

from utils import (init_data_and_model,
                   get_noise_mask,
                   add_parser_args,
                   get_2d_plot_decorators,
                   load_moments)

from argparse import ArgumentParser

use_discminer_style()

parser = ArgumentParser(prog='Identify and show peak residuals', description='Identify peak residuals in folded maps.')
parser.add_argument('-c', '--clean_thres', default=np.inf, type=float, help="Threshold above which peak residuals will be rejected.")
args = add_parser_args(parser, moment=True, kind=True, surface=True, fold=True, projection=True, Rinner=True, Router=True)

#**********************
#JSON AND PARSER STUFF
#**********************
with open('parfile.json') as json_file:
    pars = json.load(json_file)

meta = pars['metadata']
best = pars['best_fit']
custom = pars['custom']

vsys = best['velocity']['vsys']
Rout = best['intensity']['Rout']
incl = best['orientation']['incl']
PA = best['orientation']['PA']
xc = best['orientation']['xc']
yc = best['orientation']['yc']

gaps = custom['gaps']
rings = custom['rings']

ctitle, clabel, clim, cfmt, cmap_mom, cmap_res, levels_im, levels_cc, unit = get_2d_plot_decorators(args.moment, unit_simple=True, fmt_vertical=True)

#****************
#SOME DEFINITIONS
#****************
file_data = meta['file_data']
tag = meta['tag']
au_to_m = u.au.to('m')

dpc = meta['dpc']*u.pc
Rmax = 1.1*Rout*u.au #Max model radius, 10% larger than disc Rout

#*******************
#LOAD DATA AND MODEL
#*******************
datacube, model = init_data_and_model()

noise_mean, mask = get_noise_mask(datacube, thres=2)
vchannels = datacube.vchannels
model.make_model()

#*************************
#LOAD MOMENT MAPS
moment_data, moment_model, mtags = load_moments(args)

#**************************
#MASK AND COMPUTE RESIDUALS
moment_data = np.where(mask, np.nan, moment_data)
moment_model = np.where(mask, np.nan, moment_model)
moment_residuals = moment_data - moment_model
    
if args.moment=='velocity' and args.fold=='absolute':
    moment_residuals = np.abs(moment_data-vsys) - np.abs(moment_model-vsys)

#*******************
#FIND PEAK RESIDUALS
#*******************
beam_au = datacube.beam_size.to('au').value
R_prof = np.arange(args.Rinner*beam_au, args.Router*Rout, beam_au/5)
xlim0, xlim1 = 0.5*R_prof[0], 1.05*R_prof[-1]

fig, ax = plt.subplots(ncols=2, nrows=1, figsize=(11,5))    
ax_c = fig.add_axes([0.8,0.65,0.23*5/11,0.23]) #AxesSubplot for showing contour colours 

pick = Pick(model, moment_residuals, R_prof, fold=True, color_bounds = np.array([0.33, 0.66, 1.0])*Rout, ax2=ax_c)
folded_map = pick.make_2d_map() #Map where peaks will be picked from
pick.find_peaks(clean_thres=args.clean_thres)

lev = pick.lev_list
color = pick.color_list
peak_resid = pick.peak_resid
peak_angle = pick.peak_angle
peak_resid = pick.peak_resid

model.make_emission_surface(ax_c)
model.make_disc_axes(ax_c)
ax_c.axis('off')

#*******************
#SHOW PEAK RESIDUALS
#*******************
kwargs_sc = dict(facecolors=color, edgecolors='0.1', lw=0.7, alpha=0.7, zorder=10)

ax[0].set_xlabel(r'Azimuth [deg]')
ax[0].set_ylabel('Peak residual [%s]'%unit)
ax[1].set_xlabel('Radius [au]')

for axi in ax: 
    axi.tick_params(labelbottom=True, top=True, right=True, which='both', direction='in')    
    mod_major_ticks(axi, nbins=8)
    mod_minor_ticks(axi)
ax[0].set_xlim(-95,95)
ax[0].set_xticks(np.arange(-90,90+1,30))    
ax[1].tick_params(labelleft=False)

ax[0].scatter(peak_angle, peak_resid, **kwargs_sc)
ax[1].scatter(lev, peak_resid, **kwargs_sc)

ax[0].axvline(pick.peak_global_angle, lw=3, c='k', label='global peak', zorder=11)
ax[1].axvline(pick.peak_global_radius, lw=3, c='k', zorder=11) 
ax[0].legend(frameon=False, fontsize=15, handlelength=1.0, loc='lower left', bbox_to_anchor=(-0.04, 0.98))

make_substructures(ax[1], gaps=gaps, rings=rings)
append_sigma_panel(fig, ax, peak_resid, weights=pick.peak_weight, hist=True)
plt.savefig('peak_residuals_%s.png'%mtags['base'], bbox_inches='tight', dpi=200)
plt.close()

#***************
#SHOW FOLDED MAP
#***************
R = model.projected_coords['R']
phi = model.projected_coords['phi']

Xproj = R[args.surface]*np.cos(phi[args.surface])
Yproj = R[args.surface]*np.sin(phi[args.surface])

cos_peak = np.cos(np.radians(peak_angle))
sin_peak = np.sin(np.radians(peak_angle))

if args.projection=='cartesian':
    levels_resid = np.linspace(-clim, clim, 32)
    
    if args.surface in ['up', 'upper']:
        z_func = model.z_upper_func
        z_pars = best['height_upper']

    elif args.surface in ['low', 'lower']:
        z_func = model.z_lower_func
        z_pars = best['height_lower']
    
    fig, ax = make_round_map(folded_map, levels_resid, pick.X*u.au, pick.Y*u.au, R_prof[-1]*u.au,
                             z_func=z_func, z_pars=z_pars, incl=incl, PA=PA, xc=xc, yc=yc,
                             cmap=cmap_res, clabel=unit, fmt=cfmt, 
                             rings=rings,
                             mask_wedge=(90, 270)*u.deg,
                             mask_inner=R_prof[0]*u.au)
    ax.scatter(lev*cos_peak, lev*sin_peak, edgecolors='none', facecolors=color, alpha=0.2, s=100)
    ax.scatter(lev*cos_peak, lev*sin_peak, edgecolors='none', facecolors=color, alpha=1.0, s=10)
    ax.scatter(lev*cos_peak, lev*sin_peak, edgecolors='0.3', facecolors='none', alpha=1.0, s=100)
    ax.set_title('%s, folded map'%ctitle, fontsize=16, color='k')
    

plt.savefig('folded_residuals_deproj_%s_%s.png'%(mtags['base'], args.projection), bbox_inches='tight', dpi=200)
plt.close()

#*************
#FIND CLUSTERS
#*************
pick.find_clusters(n_phi=8, n_R=8)

#*************
#PLOT CLUSTERS
#*************



import os
import sys
import json
import warnings
import numpy as np
from argparse import ArgumentParser

parser = ArgumentParser(prog='make json parfile', description='Make master json file with model information to use as input for analysis scripts')
parser.add_argument('-f', '--log_file', default='', type=str, help="if none tries to guess input log file")
parser.add_argument('-o', '--overwrite', default=0, type=int, help="overwrite if parfile.json exists?")
parser.add_argument('-r', '--reset', default=0, type=int, help="If (1) and overwrite, reset 'custom' dictionary in new parfile. If not (0) and overwrite, forward 'custom' dictionary and rewrite metadata and model parameters only")
args = parser.parse_args()

#'kind' tags available: '', fixincl, I2pwl, I2pwlnosurf

#**************************
#CUSTOMISED PROPS FOR PLOTS
#**************************
#Put here any info you'd like to have access through the parfile.json in the analysis scripts 

custom_dict = {
    'chan_step': 2, #Plot channels every n steps
    'nchans': 15, #Plot n channels
    'vlim': 0.2, #Velocity max and min plot limit
    'Llim': 0.2, #Line width max and min
    'Ilim': 15, #Peak intensity max and min

    
}

#****************
#DISCS BASIC INFO
#****************
incl_dict = { #if fixincl is found in log_pars tag take inclination from mm continuum
    'mwc480': -37.0,
} #sign assumes discminer convention

gaps_dict = {
    'mwc480': [76, 149],
} #mm dust gaps

rings_dict = {
    'mwc480': [98, 165],
} #mm dust rings

kinks_dict = {
    'mwc480': [245],
} #intensity kinks

#**************************
#GET TAGS AND PARS FROM LOG 
#**************************
files_dir = os.listdir()
file_data = ''

if args.log_file == '':
    log_file = files_dir[np.argmax(['log_pars' in f for f in files_dir])]
else:
    log_file = args.log_file
    
def get_tags_dict(log_file):
    global file_data
    tag_full = log_file.split('log_pars_')[1].split('_cube')[0]
    log_file_split = np.asarray(log_file.split('_'))
    tag_disc = log_file_split[2]
    tag_mol = log_file_split[3]
    tag_dv = log_file_split[4]
    tag_program = log_file_split[5]
    icube ,= np.where(log_file_split == 'cube')[0]
    di = icube-6
    tag_kind = log_file_split[6:icube]
    if len(tag_kind)==0: tag_kind=np.asarray(['allfree'])
    tag_walkers = log_file_split[7+di].split('w')[0]
    tag_steps = log_file_split[8+di].split('s')[0]

    #Try to get filename, distance and downsampling factors from prepare_data.py
    pfile = np.asarray(os.listdir())[np.argmax(['prepare_data' in tmp for tmp in os.listdir()])] 
    prep = open(pfile, "r")
    downsamp = []
    file_read, file_clip = 0, 0
    for line in prep.readlines():
        if 'exit' in line and line[0]!='#':
            break
        if '.downsample(' in line:
            downsamp.append(line.split('(')[1].split(',')[0])
        if 'u.pc' in line:
            dpc = float(line.split('*')[0].split('=')[-1])
        if 'file_data' in line and line[0] !='#' and not file_read:
            file_read = 1
            exec(line, globals()) #file_data = 'cube.fits'; needs global vars info
        if 'datacube.clip' in line and not file_clip:
            file_clip = 1
            file_data += '_clipped'
            
    if len(downsamp)==2:
        pro, fit = np.sort(np.asarray(downsamp).astype(int))
    elif len(downsamp)==1:
        fit = int(downsamp[0])
        pro = 1
    elif len(downsamp)==0:
        pro = fit = 1
    else:
        pro = np.min(int(downsamp))
        fit = np.max(int(downsamp))
        message = "More than two downsampling factors were found in prepare_data.py script, taking smaller of all, you can specify the downsampling factor wished for prototype manually in parfile.json changing 'downsamp_pro'"
        warnings.warn(message)

    if pro > 1:
        file_data += '_downsamp_%dpix'%pro
    file_data += '.fits'

    return dict(file_data=file_data, log_file=log_file, tag=tag_full, disc=tag_disc, mol=tag_mol, dpc=dpc, dv=tag_dv, program=tag_program, kind=tag_kind.tolist(), nwalkers=int(tag_walkers), nsteps=int(tag_steps), downsamp_pro=int(pro), downsamp_fit=int(fit), downsamp_factor=(fit/pro)**2)


def get_base_pars(log_file, kind=[]):
    try:
        log_pars = np.loadtxt(log_file)[1]
        log_file_split = np.asarray(log_file.split('_'))
        tag_disc = log_file_split[2]
        
        header = np.loadtxt(log_file, comments=None, dtype=str, delimiter=' ', max_rows=1)
        header[1] = header[1][1:] #Remove first bracket of first entry
        header = [hdr[1:-2] for hdr in header[1:]] #Jump # and remove brackets and commas

        #RENAME REPEATED TAGS IN HEADER
        hdr_arr = np.array(header)
        argI0, = np.where(hdr_arr == 'I0') #Array with one entry
        argL0, = np.where(hdr_arr == 'L0')
        argLs, = np.where(hdr_arr == 'Ls')
        argz0, = np.where(hdr_arr == 'z0')
        units = []
        
        if len(argz0)==1:
            argupper, arglower = argz0[0], np.inf
            header[argupper] = 'z0_upper'
            log_pars = np.append(log_pars, np.zeros(4))
            header += ['z0_lower', 'p_lower', 'Rb_lower', 'q_lower']
        elif len(argz0)==2:
            argupper, arglower = argz0
            header[argupper] = 'z0_upper'
            header[arglower] = 'z0_lower'
        else:
            argupper, arglower = np.inf, np.inf
            log_pars = np.append(log_pars, np.zeros(8))
            header += ['z0_upper', 'p_upper', 'Rb_upper', 'q_upper', 'z0_lower', 'p_lower', 'Rb_lower', 'q_lower']

        if 'fixincl' in kind:
            log_pars = np.append(log_pars, np.radians(incl_dict[tag_disc]))
            header += ['incl']
                    
        for i in range(len(header)):
            hdr = header[i]
            if hdr in ['p', 'q', 'Rb', 'p0', 'p1']:
                diff_arg = np.array([i-argI0[0], i-argL0[0], i-argLs[0], i-argupper, i-arglower])                
                diff_arg = np.where(diff_arg<=0, np.inf, diff_arg)
                diff_min = np.argmin(diff_arg)
                if diff_min==0: header[i]+='_I0'
                if diff_min==1: header[i]+='_L0'
                if diff_min==2: header[i]+='_Ls'
                if diff_min==3: header[i]+='_upper'
                if diff_min==4: header[i]+='_lower'
                    
            #UNITS STUFF
            if 'R' in hdr or 'xc' in hdr or 'yc' in hdr or 'z' in hdr:
                units.append('au')
            elif 'incl' in hdr or 'PA' in hdr:
                units.append('rad')
            elif 'Ls' in hdr or 'p' in hdr or 'q' in hdr or 'vel_sign' in hdr:
                units.append('none')
            elif 'I0' in hdr:
                units.append('Jy/pix')
            elif 'L0' in hdr or 'vsys' in hdr:
                units.append('km/s')
            elif 'Mstar' in hdr:
                units.append('Msun')
            else:
                units.append('unknown')
                        
        #********
        #SIMPLE DICT OF BEST FIT PARS AND UNITS
        par_dict = dict(zip(header, log_pars))
        uni_dict = dict(zip(header, units))        

        #********
        #CLASSIFY PARS IN ATTRIBUTES
        par_dict_att = {
            'velocity': {}, 'orientation': {},
            'intensity': {}, 'linewidth': {}, 'lineslope': {},
            'height_upper': {}, 'height_lower': {}
        }
        uni_dict_att = {
            'velocity': {}, 'orientation': {},
            'intensity': {}, 'linewidth': {}, 'lineslope': {},
            'height_upper': {}, 'height_lower': {}
        }

        def update_dict(att, key):
            if key=='vel_sign':
                key_sp = key
            else:
                key_sp = key.split('_')[0]
            par_dict_att[att].update({key_sp: par_dict[key]})
            uni_dict_att[att].update({key_sp: uni_dict[key]})            
            
        for key in par_dict:
            if key in ['Mstar', 'vsys', 'vel_sign']:
                update_dict('velocity', key)
            if key in ['incl', 'PA', 'xc', 'yc']:
                update_dict('orientation', key)
            if 'I0' in key or 'Rout' in key or 'Rbreak' in key:
                update_dict('intensity', key)
            if 'L0' in key:
                update_dict('linewidth', key)
            if 'Ls' in key:
                update_dict('lineslope', key)
            if 'upper' in key:
                update_dict('height_upper', key)
            if 'lower' in key:
                update_dict('height_lower', key)
                
        return par_dict_att, uni_dict_att
    
    except FileNotFoundError:
        message = 'Log par file not found: %s'%log_file
        warnings.warn(message)
        return None, None

#*************
#MAKE PAR FILE
#*************    
def make_json(dicts_list=[], keys_list=[], filename='parfile.json'):
    master_dict = {key: dicts_list[i] for i,key in enumerate(keys_list)}

    def make():
        with open(filename, 'w', encoding='utf-8') as f:
            json.dump(master_dict, f, ensure_ascii=False, indent=4)

    if args.overwrite:
        if 'parfile.json' in os.listdir():
            message = 'Overwriting JSON parfile...'
            #warnings.warn(message)
            print ('*'*10 + message)
            if args.reset:
                message = "'custom' dictionary was reset to default values..."
                warnings.warn(message)                

            if not args.reset:
                with open('parfile.json') as json_file:
                    pars = json.load(json_file)
                message = "Forwarding 'custom' dictionary to the new parfile.json..."
                print ('*'*10 + message)
                    
                master_dict['custom'].update(pars['custom'])
        make()
        
    else:
        if 'parfile.json' in os.listdir():
            message = '\n********JSON parfile exists and overwrite mode is off, therefore NO output was produced********'
            warnings.warn(message)
        else:
            make()

if __name__ == '__main__':
    tags_dict = get_tags_dict(log_file)
    pars_dict, units_dict = get_base_pars(log_file, kind=tags_dict['kind'])
    for tk in tags_dict['kind']:
        if 'nosurf' in tk:
            pars_dict['intensity'].update({'q': 0.0})
            units_dict['intensity'].update({'q': "none"})            
            pars_dict['linewidth'].update({'q': 0.0})
            units_dict['linewidth'].update({'q': "none"})
            
    custom_dict.update(gaps=gaps_dict[tags_dict['disc']], rings=rings_dict[tags_dict['disc']], kinks=kinks_dict[tags_dict['disc']]) 
    make_json(dicts_list = [custom_dict, tags_dict, pars_dict, units_dict], keys_list = ['custom', 'metadata', 'best_fit', 'units'])
